import numpy as np
import pytest

from dclab.features import emodulus


@pytest.mark.filterwarnings('ignore::dclab.features.emodulus.'
                            + 'YoungsModulusLookupTableExceededWarning')
def test_simple_emod_he_3d_fem_22():
    x = np.linspace(0, 250, 100)
    y = np.linspace(0, 0.1, 100)
    x, y = np.meshgrid(x, y)

    emod = emodulus.get_emodulus(area_um=x,
                                 deform=y,
                                 medium="CellCarrier",
                                 channel_width=30,
                                 flow_rate=0.16,
                                 lut_data="HE-3D-FEM-22",
                                 px_um=0,  # without pixelation correction
                                 temperature=23,
                                 visc_model="herold-2017")

    assert np.allclose(emod[10, 50], 1.301024899951213)
    assert np.allclose(emod[50, 50], 0.6713689164185259)
    assert np.isnan(emod[80, 50])

    assert np.allclose(emod[10, 80], 1.7172087155883002)
    assert np.allclose(emod[50, 80], 0.8595551172406175)
    assert np.isnan(emod[80, 80])



import warnings


class MachineLearningFeature:
    def __init__(self, *args, **kwargs):
        raise NotImplementedError(
            "`MachineLearningFeature` has been stripped from dclab since "
            "version 0.61.0.")


def load_ml_feature(*args, **kwargs):
    raise NotImplementedError(
        "`load_ml_feature` has been stripped from dclab since "
        "version 0.61.0.")


def load_modc(*args, **kwargs):
    raise NotImplementedError(
        "`load_modc` has been stripped from dclab since "
        "version 0.61.0.")


def remove_all_ml_features():
    warnings.warn("The `remove_all_ml_features` does nothing since it was "
                  "stripped from dclab version 0.61.0.",
                  DeprecationWarning)


def save_modc(*args, **kwargs):
    raise NotImplementedError(
        "`save_modc` has been stripped from dclab since "
        "version 0.61.0.")



"""Repack (similar to h5repack) .rtdc files"""
from __future__ import annotations

import argparse
import pathlib

import h5py

from ..rtdc_dataset import rtdc_copy
from .._version import version

from . import common


def repack(
        path_in: str | pathlib.Path = None,
        path_out: str | pathlib.Path = None,
        strip_basins: bool = False,
        strip_logs: bool = False,
        check_suffix: bool = True,
        ret_path: bool = False,
        ):
    """Repack/recreate an .rtdc file, optionally stripping the logs

    Parameters
    ----------
    path_in: str or pathlib.Path
        file to compress
    path_out: str or pathlib
        output file path
    strip_basins: bool
        do not write basin information to the output file
    strip_logs: bool
        do not write logs to the output file
    check_suffix: bool
        check suffixes for input and output files
    ret_path: bool
        whether to return the output path

    Returns
    -------
    path_out: pathlib.Path
        output path (with possibly corrected suffix)
    """
    if path_in is None and path_out is None:
        parser = repack_parser()
        args = parser.parse_args()
        path_in = args.input
        path_out = args.output
        strip_basins = args.strip_basins
        strip_logs = args.strip_logs

    allowed_input_suffixes = [".rtdc"]
    if not check_suffix:
        allowed_input_suffixes.append(pathlib.Path(path_in).suffix)

    path_in, path_out, path_temp = common.setup_task_paths(
        path_in, path_out, allowed_input_suffixes=allowed_input_suffixes)

    with h5py.File(path_in) as h5, h5py.File(path_temp, "w") as hc:
        rtdc_copy(src_h5file=h5,
                  dst_h5file=hc,
                  features="all",
                  include_basins=not strip_basins,
                  include_logs=not strip_logs,
                  include_tables=True,
                  meta_prefix="")

    # Finally, rename temp to out
    path_temp.rename(path_out)
    if ret_path:
        return path_out


def repack_parser():
    descr = "Repack an .rtdc file. The difference to dclab-compress " \
            + "is that no logs are added. Other logs can optionally be " \
            + "stripped away. Repacking also gets rid of old clutter " \
            + "data (e.g. previous metadata stored in the HDF5 file)."
    parser = argparse.ArgumentParser(description=descr)
    parser.add_argument('input', metavar="INPUT", type=str,
                        help='Input path (.rtdc file)')
    parser.add_argument('output',  metavar="OUTPUT", type=str,
                        help='Output path (.rtdc file)')
    parser.add_argument('--strip-basins',
                        dest='strip_basins',
                        action='store_true',
                        help='Do not copy any basin information to the '
                             'output file.')
    parser.set_defaults(strip_basins=False)
    parser.add_argument('--strip-logs',
                        dest='strip_logs',
                        action='store_true',
                        help='Do not copy any logs to the output file.')
    parser.set_defaults(strip_logs=False)
    parser.add_argument('--version', action='version',
                        version=f'dclab-repack {version}')
    return parser



"""Test CLI dclab-join"""
import shutil

import dclab
from dclab import cli, new_dataset, rtdc_dataset, RTDCWriter

import h5py
import numpy as np
import pytest

from helper_methods import retrieve_data


def test_join_tdms():
    pytest.importorskip("nptdms")
    path_in = retrieve_data("fmt-tdms_shapein-2.0.1-no-image_2017.zip")
    # same directory (will be cleaned up with path_in)
    path_out = path_in.with_name("out.rtdc")

    cli.join(paths_in=[path_in, path_in], path_out=path_out)

    with new_dataset(path_out) as dsj, new_dataset(path_in) as ds0:
        assert len(dsj)
        assert len(dsj) == 2 * len(ds0)
        assert len(ds0) == ds0.config["experiment"]["event count"]
        assert len(dsj) == dsj.config["experiment"]["event count"]
        assert np.all(dsj["circ"][:100] == ds0["circ"][:100])
        assert np.all(dsj["circ"][len(ds0):len(ds0)+100] == ds0["circ"][:100])
        assert set(dsj.features) == set(ds0.features)


def test_join_tdms_logs():
    pytest.importorskip("nptdms")
    path_in = retrieve_data("fmt-tdms_shapein-2.0.1-no-image_2017.zip")
    # same directory (will be cleaned up with path_in)
    path_out = path_in.with_name("out.rtdc")

    cli.join(paths_in=[path_in, path_in], path_out=path_out)

    with new_dataset(path_out) as dsj, new_dataset(path_in) as ds0:
        assert "dclab-join" in dsj.logs
        assert "src-#1_cfg" in dsj.logs
        assert "software version = ShapeIn 2.0.1" in dsj.logs["src-#1_cfg"]
        assert "software version = ShapeIn 2.0.1" in dsj.logs["src-#2_cfg"]
        assert "src-#1_M1_camera.ini" in dsj.logs
        assert "src-#2_M1_camera.ini" in dsj.logs
        assert ds0.logs
        for key in ds0.logs:
            jkey = "src-#1_" + key
            assert np.all(np.array(ds0.logs[key]) == np.array(dsj.logs[jkey]))


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_join_rtdc():
    path_in = retrieve_data("fmt-hdf5_mask-contour_2018.zip")
    # same directory (will be cleaned up with path_in)
    path_out = path_in.with_name("out.rtdc")

    ret = cli.join(paths_in=[path_in, path_in], path_out=path_out)
    assert ret is None, "by default, this method should return 0 (exit 0)"
    with new_dataset(path_out) as dsj, new_dataset(path_in) as ds0:
        assert "dclab-join" in dsj.logs
        assert len(dsj)
        assert len(dsj) == 2 * len(ds0)
        assert np.all(dsj["circ"][:len(ds0)] == ds0["circ"])
        assert np.all(dsj["circ"][len(ds0):] == ds0["circ"])
        assert set(dsj.features) == set(ds0.features)
        assert 'identifier = ZMDD-AcC-8ecba5-cd57e2' in dsj.logs["src-#1_cfg"]
        assert 'identifier = ZMDD-AcC-8ecba5-cd57e2' in dsj.logs["src-#2_cfg"]


def test_join_rtdc_basin_data_not_written():
    """When joining data, basin features should not be written"""
    h5path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")
    h5path_small = h5path.with_name("smaller.rtdc")
    h5path_out = h5path.with_name("compressed.rtdc")

    # Dataset creation
    with h5py.File(h5path) as src, RTDCWriter(h5path_small) as hw:
        # first, copy all the scalar features to the new file
        rtdc_dataset.rtdc_copy(src_h5file=src,
                               dst_h5file=hw.h5file,
                               features=["deform", "area_um"])
        hw.store_basin(basin_name="example basin",
                       basin_type="file",
                       basin_format="hdf5",
                       basin_locs=[h5path],
                       basin_descr="an example test basin",
                       )

    # sanity check
    with new_dataset(h5path_small) as ds:
        assert "image" in ds.features

    # compress the basin-based dataset
    cli.join(paths_in=[h5path_small, h5path_small], path_out=h5path_out)

    with h5py.File(h5path_out) as h5:
        assert "basins" not in h5

    with new_dataset(h5path_out) as ds:
        assert "image" not in ds
        assert "aspect" not in ds
        assert "deform" in ds


def test_join_rtdc_basin_information_lost():
    """When joining .rtdc files, the basin information is lost

    Or at least this is not implemented in dclab 0.58.0.
    """
    h5path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")
    h5path_small = h5path.with_name("smaller.rtdc")
    h5path_out = h5path.with_name("compressed.rtdc")

    # Dataset creation
    with h5py.File(h5path) as src, RTDCWriter(h5path_small) as hw:
        # first, copy all the scalar features to the new file
        rtdc_dataset.rtdc_copy(src_h5file=src,
                               dst_h5file=hw.h5file,
                               features="scalar")
        hw.store_basin(basin_name="example basin",
                       basin_type="file",
                       basin_format="hdf5",
                       basin_locs=[h5path],
                       basin_descr="an example test basin",
                       )

    # sanity check
    with new_dataset(h5path_small) as ds:
        assert "image" in ds.features

    # compress the basin-based dataset
    cli.join(paths_in=[h5path_small, h5path_small], path_out=h5path_out)

    with h5py.File(h5path_out) as h5:
        assert "basins" not in h5


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_join_rtdc_logs():
    path_in = retrieve_data("fmt-hdf5_mask-contour_2018.zip")
    # create second file with new log
    path_in_2 = path_in.with_name("second.rtdc")
    shutil.copy2(path_in, path_in_2)
    with dclab.RTDCWriter(path_in_2, mode="append") as hw:
        hw.store_log("Dummy", [
            "One, two, three, four",
            "Everybody get on the dance floor",
            "Five, six, seven, eight",
            "It's 4 a.m. and I'm wide awake"
        ])
    # same directory (will be cleaned up with path_in)
    path_out = path_in.with_name("out.rtdc")

    cli.join(paths_in=[path_in, path_in_2], path_out=path_out)

    with h5py.File(path_out) as h5:
        assert "src-#2_Dummy" in h5["logs"]


def test_join_rtdc_index_online_issue_158():
    """
    dclab did not correctly access events/index_online before
    """
    path1 = retrieve_data("fmt-hdf5_polygon_gate_2021.zip")
    path2 = retrieve_data("fmt-hdf5_polygon_gate_2021.zip")
    path_out_a = path1.with_name("outa.rtdc")

    # this did not work
    cli.join(paths_in=[path1, path2], path_out=path_out_a)

    # verification
    with dclab.new_dataset(path_out_a) as ds:
        assert "index_online" in ds.features_innate
        assert np.all(np.diff(ds["index_online"]) > 0)


def test_join_rtdc_unequal_features_issue_157():
    """
    If two files do not contain the same number of features, joining
    should only take into account the same features and then should
    issue a warning (which will then be written to the logs in DCKit).
    """
    path1 = retrieve_data("fmt-hdf5_polygon_gate_2021.zip")
    path2 = retrieve_data("fmt-hdf5_polygon_gate_2021.zip")
    path_out_a = path1.with_name("outa.rtdc")
    path_out_b = path1.with_name("outb.rtdc")
    path_out_c = path1.with_name("outc.rtdc")

    # add a feature to path1 and define order (join sorts the files with date)
    with h5py.File(path1, "a") as h51:
        h51["events"]["volume"] = h51["events"]["area_um"][:] ** 1.5
        h51.attrs["experiment:date"] = "2020-01-01"
    with h5py.File(path2, "a") as h51:
        h51.attrs["experiment:date"] = "2021-01-01"

    # sanity checks
    with dclab.new_dataset(path1) as ds:
        assert "volume" in ds.features_innate
    with dclab.new_dataset(path2) as ds:
        assert "volume" not in ds.features_innate

    # First test
    # There should be no warning here, because for path2 volume can be
    # computed.
    cli.join(paths_in=[path1, path2], path_out=path_out_a)
    with dclab.new_dataset(path_out_a) as ds:
        # Volume is in this file, because it can be computed for path2
        assert "volume" in ds.features_innate
        # this can only be mapped with uint64
        assert ds["frame"][-1] == 632448000038

    # Second test: Now do the same thing with reversed dates
    with h5py.File(path2, "a") as h51:
        h51.attrs["experiment:date"] = "2019-01-01"
    cli.join(paths_in=[path1, path2], path_out=path_out_b)
    with dclab.new_dataset(path_out_b) as ds:
        assert "volume" not in ds.features_innate
        assert "dclab-join-feature-warnings" in ds.logs
        log = "\n".join(ds.logs["dclab-join-feature-warnings"])
        assert "volume" in log

    # Third test: we flip around paths_in to also test sorting
    with h5py.File(path2, "a") as h51:
        h51.attrs["experiment:date"] = "2019-01-01"
    cli.join(paths_in=[path2, path1], path_out=path_out_c)
    with dclab.new_dataset(path_out_c) as ds:
        assert "volume" not in ds.features_innate
        assert "dclab-join-feature-warnings" in ds.logs
        log = "\n".join(ds.logs["dclab-join-feature-warnings"])
        assert "volume" in log


def test_join_rtdc_unequal_features_issue_157_2():
    """
    Same test as above, but we use a feature that cannot be computed
    """
    path1 = retrieve_data("fmt-hdf5_polygon_gate_2021.zip")
    path2 = retrieve_data("fmt-hdf5_polygon_gate_2021.zip")
    path_out_a = path1.with_name("outa.rtdc")
    path_out_b = path1.with_name("outb.rtdc")

    # add a feature to path1 and define order (join sorts the files with date)
    with h5py.File(path1, "a") as h51:
        h51["events"]["ml_score_abc"] = np.log10(h51["events"]["area_um"][:])
        h51.attrs["experiment:date"] = "2020-01-01"
    with h5py.File(path2, "a") as h51:
        h51.attrs["experiment:date"] = "2021-01-01"

    # sanity checks
    with dclab.new_dataset(path1) as ds:
        assert "ml_score_abc" in ds.features_innate
    with dclab.new_dataset(path2) as ds:
        assert "ml_score_abc" not in ds.features_innate

    # First test
    cli.join(paths_in=[path1, path2], path_out=path_out_a)
    with dclab.new_dataset(path_out_a) as ds:
        # Score cannot be computed for path2
        assert "ml_score_abc" not in ds.features_innate
        assert "dclab-join-feature-warnings" in ds.logs
        log = "\n".join(ds.logs["dclab-join-feature-warnings"])
        assert "ml_score_abc" in log

    # Second test: Now do the same thing with reversed dates
    with h5py.File(path2, "a") as h51:
        h51.attrs["experiment:date"] = "2019-01-01"
    cli.join(paths_in=[path1, path2], path_out=path_out_b)
    with dclab.new_dataset(path_out_b) as ds:
        assert "ml_score_abc" not in ds.features_innate
        assert "dclab-join-feature-warnings" in ds.logs
        log = "\n".join(ds.logs["dclab-join-feature-warnings"])
        assert "ml_score_abc" in log


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_join_frame():
    path_in1 = retrieve_data("fmt-hdf5_mask-contour_2018.zip")
    path_in2 = retrieve_data("fmt-hdf5_mask-contour_2018.zip")
    # same directory (will be cleaned up with path_in)
    path_out = path_in1.with_name("out.rtdc")

    # modify acquisition times
    with h5py.File(path_in1, mode="a") as h1:
        h1.attrs["experiment:date"] = "2019-11-04"
        h1.attrs["experiment:time"] = "15:00:00"

    with h5py.File(path_in2, mode="a") as h2:
        h2.attrs["experiment:date"] = "2019-11-05"
        h2.attrs["experiment:time"] = "16:01:15.050"

    offset = 24 * 60 * 60 + 60 * 60 + 1 * 60 + 15 + .05

    cli.join(paths_in=[path_in1, path_in2], path_out=path_out)
    with new_dataset(path_out) as dsj, new_dataset(path_in1) as ds0:
        fr = ds0.config["imaging"]["frame rate"]
        frame0 = np.array(ds0["frame"], dtype=int)
        assert np.allclose(dsj["frame"],
                           np.concatenate((frame0,
                                           frame0 + int(round(offset * fr)))),
                           rtol=0,
                           atol=0)
        assert dsj["frame"][-1] == 360302210


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_join_times():
    path_in1 = retrieve_data("fmt-hdf5_mask-contour_2018.zip")
    path_in2 = retrieve_data("fmt-hdf5_mask-contour_2018.zip")
    # same directory (will be cleaned up with path_in)
    path_out = path_in1.with_name("out.rtdc")

    # modify acquisition times
    with h5py.File(path_in1, mode="a") as h1:
        h1.attrs["experiment:date"] = "2019-11-04"
        h1.attrs["experiment:time"] = "15:00:00"
        # Necessary due to issue #204
        time = np.array(h1["events/time"], dtype=float)
        del h1["events/time"]
        h1["events/time"] = time

    with h5py.File(path_in2, mode="a") as h2:
        h2.attrs["experiment:date"] = "2019-11-05"
        h2.attrs["experiment:time"] = "16:01:15.050"
        # Necessary due to issue #204
        time = np.array(h2["events/time"], dtype=float)
        del h2["events/time"]
        h2["events/time"] = time

    offset = 24 * 60 * 60 + 60 * 60 + 1 * 60 + 15 + .05

    cli.join(paths_in=[path_in1, path_in2], path_out=path_out)
    with new_dataset(path_out) as dsj, new_dataset(path_in1) as ds0:
        assert np.allclose(dsj["time"],
                           np.concatenate((ds0["time"], ds0["time"] + offset)),
                           rtol=0,
                           atol=.0001)



"""Plotting custom isoelastics

This example illustrates how to extract custom isoelasticity lines from
the dclab look-up tables by reproducing figure 3 (right) of
:cite:`Wittwer2022`.

Note that at the boundary of the support of a look-up table, the isoelasticity
lines may break away in perpendicular directions. The underlying reason
is that the look-up table is first mapped onto a grid from which the
constant isoelasticity lines are extracted. Since the Young's modulus values
are linearly interpolated from the LUT onto that grid, there can be
inaccuracies for pixels that are at the LUT boundary.

An elaborate way of getting rid of these inaccuracies (and this is
how the isoelasticity lines for dclab are extracted), is to extend
the LUT by fitting a polynomial to isoelasticity lines which are well-defined
within the LUT and extrapolating these lines beyond the boundary of the LUT.
This technique is documented in the `scripts` directory of the dclab
repository.

A quicker and much less elaborate way of getting around this
issue is to simply crop the individual isoelasticity lines where necessary.
"""
import matplotlib.pylab as plt
import numpy as np
import skimage

import dclab
from dclab.features import emodulus


colors = ["r", "b"]
linestyles = [":", "-"]

plt.figure(figsize=(8, 4))
ax = plt.subplot(111,
                 title="Comparison of the isoelasticity lines of two LUTs")

grid_sie = 250

for ii, lut_name in enumerate(["LE-2D-FEM-19", "HE-3D-FEM-22"]):
    area_um = np.linspace(0, 350, grid_sie, endpoint=True)
    deform = np.linspace(0, 0.2, grid_sie, endpoint=True)
    area_um_grid, deform_grid = np.meshgrid(area_um, deform, indexing="ij")

    emod = emodulus.get_emodulus(area_um=area_um_grid,
                                 deform=deform_grid,
                                 medium=6.0,
                                 channel_width=20,
                                 flow_rate=0.04,
                                 px_um=0,
                                 temperature=None,
                                 visc_model=None,
                                 lut_data=lut_name)

    levels = [0.5, 0.75, 1.0, 1.25, 1.5, 2.0, 3.0, 6.0]
    for level in levels:
        conts = skimage.measure.find_contours(emod, level=level)
        if not conts:
            continue
        # get the longest one
        idx = np.argmax([len(cc) for cc in conts])
        cc = conts[idx]
        # remove nan values
        cc = cc[~np.isnan(np.sum(cc, axis=1))]
        # scale isoelastics back
        cc_sc = np.copy(cc)
        cc_sc[:, 0] = cc[:, 0] / grid_sie * 350
        cc_sc[:, 1] = cc[:, 1] / grid_sie * 0.2
        plt.plot(cc_sc[:, 0], cc_sc[:, 1],
                 color=colors[ii],
                 ls=linestyles[ii],
                 label=lut_name if level == levels[0] else None)

ax.set_ylim(-0.005, 0.1)
ax.set_xlabel(dclab.dfn.get_feature_label("area_um"))
ax.set_ylabel(dclab.dfn.get_feature_label("deform"))
plt.legend()
plt.tight_layout()
plt.show()



"""
Computation of the 10th and 90th percentile of grayscale values inside the
RT-DC event image mask with background-correction taken into account.
"""
import numpy as np


def get_bright_perc(mask, image, image_bg, bg_off=None):
    """Compute 10th and 90th percentile of the bg-corrected event brightness

    The background-corrected event brightness is defined by the
    gray-scale values of the background-corrected image data
    within the event mask area.

    Parameters
    ----------
    mask: ndarray or list of ndarrays of shape (M,N) and dtype bool
        The mask values, True where the event is located in `image`.
    image: ndarray or list of ndarrays of shape (M,N)
        A 2D array that holds the image in form of grayscale values
        of an event.
    image_bg: ndarray or list of ndarrays of shape (M,N)
        A 2D array that holds the background image for the same event.
    bg_off: float or 1D ndarray
        Additional offset value that is added to `image_bg` before
        background correction

    Returns
    -------
    bright_perc_10: float or ndarray of size N
        10th percentile of brightness
    bright_perc_10: float or ndarray of size N
        90th percentile of brightness
    """
    if isinstance(mask, np.ndarray) and len(mask.shape) == 2:
        # We have a single image
        image_bg = [image_bg]
        image = [image]
        mask = [mask]
        ret_list = False
    else:
        ret_list = True

    length = min(len(mask), len(image), len(image_bg))

    p10 = np.zeros(length, dtype=np.float64) * np.nan
    p90 = np.zeros(length, dtype=np.float64) * np.nan

    for ii in range(length):
        # cast to integer before subtraction
        imgi = np.array(image[ii], dtype=int) - image_bg[ii]
        mski = mask[ii]
        # Assign results
        p10[ii], p90[ii] = np.percentile(imgi[mski], q=[10, 90])

    if bg_off:
        p10 -= bg_off
        p90 -= bg_off

    if ret_list:
        return p10, p90
    else:
        return p10[0], p90[0]



"""Utility methods"""
import functools
import hashlib
import numbers
import pathlib
import warnings

import h5py
import numpy as np
from .rtdc_dataset.config import Configuration, ConfigurationDict


if np.lib.NumpyVersion(np.__version__) >= "2.0.0":
    copy_if_needed = None
else:
    copy_if_needed = False


class file_monitoring_lru_cache:
    """Decorator for caching data extracted from files

    The function that is decorated with `file_monitoring_lru_cache`
    must accept `path` as its first argument. Caching is
    done with an `lru_cache`. In addition to the full path
    and the other arguments to the decorated function, the
    size and the modification time of `path` is used as a
    key for the decorator.
    If the path does not exist, no caching is done.

    Use case: Extract and cache metadata from a file on disk
    that may change.
    """
    def __init__(self, maxsize=100):
        self.lru_cache = functools.lru_cache(maxsize=maxsize)
        self.cached_wrapper = None

    def __call__(self, func):
        @self.lru_cache
        def cached_wrapper(path, path_stats, *args, **kwargs):
            assert path_stats, "We need stat for validating the cache"
            return func(path, *args, **kwargs)

        @functools.wraps(func)
        def wrapper(path, *args, **kwargs):
            full_path = pathlib.Path(path).resolve()
            if full_path.exists():
                path_stat = full_path.stat()
                return cached_wrapper(
                    path=full_path,
                    path_stats=(path_stat.st_mtime_ns, path_stat.st_size),
                    *args,
                    **kwargs)
            else:
                # `func` will most-likely raise an exception
                return func(path, *args, **kwargs)

        wrapper.cache_clear = cached_wrapper.cache_clear
        wrapper.cache_info = cached_wrapper.cache_info

        return wrapper


@file_monitoring_lru_cache(maxsize=100)
def hashfile(fname, blocksize=65536, count=0, constructor=hashlib.md5,
             hasher_class=None):
    """Compute md5 hex-hash of a file

    Parameters
    ----------
    fname: str or pathlib.Path
        path to the file
    blocksize: int
        block size in bytes read from the file
        (set to `0` to hash the entire file)
    count: int
        number of blocks read from the file
    hasher_class: callable
        deprecated, see use `constructor` instead
    constructor: callable
        hash algorithm constructor
    """
    if hasher_class is not None:
        warnings.warn("The `hasher_class` argument is deprecated, please use "
                      "`constructor` instead.")
        constructor = hasher_class

    path = pathlib.Path(fname)

    hasher = constructor()
    with path.open('rb') as fd:
        buf = fd.read(blocksize)
        ii = 0
        while len(buf) > 0:
            hasher.update(buf)
            buf = fd.read(blocksize)
            ii += 1
            if count and ii == count:
                break
    return hasher.hexdigest()


def hashobj(obj):
    """Compute md5 hex-hash of a Python object"""
    return hashlib.md5(obj2bytes(obj)).hexdigest()


def obj2bytes(obj):
    """Bytes representation of an object for hashing

    Note that there is no guarantee that the bytes representation
    returned is reproducible across sessions. This is currently the
    case when an :class:`.RTDCBase` instance is passed. There is no
    opinion on wether/how this should be changed.
    """
    if isinstance(obj, str):
        return obj.encode("utf-8")
    elif isinstance(obj, pathlib.Path):
        return obj2bytes(str(obj))
    elif isinstance(obj, (bool, numbers.Number)):
        return str(obj).encode("utf-8")
    elif obj is None:
        return b"none"
    elif isinstance(obj, np.ndarray):
        return obj.tobytes()
    elif isinstance(obj, tuple):
        return obj2bytes(list(obj))
    elif isinstance(obj, list):
        return b"".join(obj2bytes(o) for o in obj)
    elif isinstance(obj, dict):
        return obj2bytes(sorted(obj.items()))
    elif hasattr(obj, "identifier"):
        # For RTDCBase, this identifier is not reproducible in-between
        # sessions. We might want to change this to something that is
        # reproducible in the future (if the need arises).
        return obj2bytes(obj.identifier)
    elif isinstance(obj, h5py.Dataset):
        # path within the HDF5 file
        o_name = obj.name
        # filename
        o_filename = obj.file.filename
        _data = [o_name, o_filename]
        if pathlib.Path(o_filename).exists():
            # when the file was changed
            _data.append(pathlib.Path(obj.file.filename).stat().st_mtime)
            # size of the file
            _data.append(pathlib.Path(obj.file.filename).stat().st_size)
        return obj2bytes(_data)
    elif hasattr(obj, "__array__"):  # must come after h5py.Dataset
        return obj2bytes(obj.__array__())
    elif isinstance(obj, Configuration):
        return obj2bytes(obj.tostring())
    elif isinstance(obj, ConfigurationDict):
        return obj2bytes(dict(obj))
    else:
        raise ValueError("No rule to convert object '{}' to string.".
                         format(obj.__class__))



"""Viscosity computation for various media"""
from __future__ import annotations

from typing import Literal
import warnings

import numpy as np

from ...warn import PipelineWarning


#: Dictionary with different names for one medium
SAME_MEDIA = {
    "0.49% MC-PBS": ["0.49% MC-PBS",
                     "0.5% MC-PBS",
                     "0.50% MC-PBS",
                     "CellCarrier",
                     ],
    # 6g MC in 1010g PBS [!sic]
    "0.59% MC-PBS": ["0.59% MC-PBS",
                     "0.6% MC-PBS",
                     "0.60% MC-PBS",
                     "CellCarrier B",
                     "CellCarrierB",
                     ],
    "0.83% MC-PBS": ["0.83% MC-PBS",
                     "0.8% MC-PBS",
                     "0.80% MC-PBS"],
    "water": ["water"],
}

#: Many media names are actually shorthand for one medium
ALIAS_MEDIA = {}
for key in SAME_MEDIA:
    for item in SAME_MEDIA[key]:
        ALIAS_MEDIA[item] = key
        ALIAS_MEDIA[item.lower()] = key  # also support all-lower case

#: Media for which computation of viscosity is defined (has duplicate entries)
KNOWN_MEDIA = sorted(ALIAS_MEDIA.keys())


class TemperatureOutOfRangeWarning(PipelineWarning):
    pass


def check_temperature(model: str,
                      temperature: float | np.array,
                      tmin: float,
                      tmax: float):
    """Raise a TemperatureOutOfRangeWarning if applicable"""
    if np.min(temperature) < tmin or np.max(temperature) > tmax:
        warnings.warn(
            f"For the {model} model, the temperature should be "
            f"in [{tmin}, {tmax}] degC! Got min/max of "
            f"[{np.min(temperature):.1f}, {np.max(temperature):.1f}] degC.",
            TemperatureOutOfRangeWarning)


def get_viscosity(medium: str = "0.49% MC-PBS",
                  channel_width: float = 20.0,
                  flow_rate: float = 0.16,
                  temperature: float | np.ndarray = 23.0,
                  model: Literal['herold-2017',
                                 'herold-2017-fallback',
                                 'buyukurganci-2022',
                                 'kestin-1978'] = 'herold-2017-fallback'):
    """Returns the viscosity for RT-DC-specific media

    Media that are not pure (e.g. ketchup or polymer solutions)
    often exhibit a non-linear relationship between shear rate
    (determined by the velocity profile) and shear stress
    (determined by pressure differences). If the shear stress
    grows non-linearly with the shear rate resulting in a slope
    in log-log space that is less than one, then we are talking about
    shear thinning. The viscosity is not a constant anymore (as it
    is e.g. for water). At higher flow rates, the viscosity becomes
    smaller, following a power law. Christoph Herold characterized
    shear thinning for the CellCarrier media :cite:`Herold2017`.
    The resulting formulae for computing the viscosities of these
    media at different channel widths, flow rates, and temperatures,
    are implemented here.

    Parameters
    ----------
    medium: str
        The medium to compute the viscosity for; Valid values
        are defined in :const:`KNOWN_MEDIA`.
    channel_width: float
        The channel width in µm
    flow_rate: float
        Flow rate in µL/s
    temperature: float or ndarray
        Temperature in °C
    model: str
        The model name to use for computing the medium viscosity.
        For water, this value is ignored, as there is only the
        'kestin-1978' model :cite:`Kestin_1978`. For MC-PBS media,
        there are the 'herold-2017' model :cite:`Herold2017` and the
        'buyukurganci-2022' model :cite:`Buyukurganci2022`.

    Returns
    -------
    viscosity: float or ndarray
        Viscosity in mPa*s

    Notes
    -----
    - CellCarrier (0.49% MC-PBS) and CellCarrier B (0.59% MC-PBS) are
      media designed for RT-DC experiments.
    - A :class:`TemperatureOutOfRangeWarning` is issued if the
      input temperature range exceeds the temperature ranges of
      the models.
    """
    # also support lower-case media and a space before the "B"
    if medium not in KNOWN_MEDIA:
        raise ValueError(f"Invalid medium: {medium}")
    medium = ALIAS_MEDIA[medium]

    if medium == "water":
        # We ignore the `model`, because it's user convenient.
        eta = get_viscosity_water_kestin_1978(temperature=temperature)
    elif not flow_rate:
        # When flow rate is zero, we cannot compute the viscosity for
        # anything other than water above.
        eta = np.nan
    elif medium in ["0.49% MC-PBS", "0.59% MC-PBS", "0.83% MC-PBS"]:
        kwargs = {"medium": medium,
                  "temperature": temperature,
                  "flow_rate": flow_rate,
                  "channel_width": channel_width}

        # Let the user know that we have a new model in town.
        if model == "herold-2017-fallback":
            warnings.warn(
                "dclab 0.48.0 introduced a more accurate model for computing "
                "the MC-PBS viscosity. You are now using the old model "
                "'herold-2017'. Unless you are reproducing an old analysis "
                "pipeline, you should consider passing 'buyukurganci-2022' "
                "as a viscosity model!",
                DeprecationWarning)
            model = "herold-2017"

        if model == "herold-2017":
            eta = get_viscosity_mc_pbs_herold_2017(**kwargs)
        elif model == "buyukurganci-2022":
            eta = get_viscosity_mc_pbs_buyukurganci_2022(**kwargs)
        else:
            raise NotImplementedError(f"Unknown model '{model}' for MC-PBS!")
    else:
        raise NotImplementedError(f"Unknown medium '{medium}'!")
    return eta


def shear_rate_square_channel(flow_rate, channel_width, flow_index):
    """Returns The wall shear rate of a power law liquid in a squared channel.

    Parameters
    ----------
    flow_rate: float
        Flow rate in µL/s
    channel_width: float
        The channel width in µm
    flow_index: float
        Flow behavior index aka the power law exponent of the shear thinning

    Returns
    -------
    shear_rate: float
        Shear rate in 1/s.
    """
    # convert channel width to mm
    channel_width = channel_width * 1e-3
    return 8*flow_rate/(channel_width**3) * (0.6671 + 0.2121/flow_index)


def get_viscosity_mc_pbs_buyukurganci_2022(
        medium: Literal["0.49% MC-PBS",
                        "0.59% MC-PBS",
                        "0.83% MC-PBS"] = "0.49% MC-PBS",
        channel_width: float = 20.0,
        flow_rate: float = 0.16,
        temperature: float = 23.0):
    """Compute viscosity of MC-PBS according to :cite:`Buyukurganci2022`

    This viscosity model was derived in :cite:`Buyukurganci2022`
    and adapted for RT-DC in :cite:`Reichel2023`.
    """
    check_temperature("'buyukurganci-2022' MC-PBS", temperature, 22, 37)
    # material constants for temperature behavior of MC dissolved in PBS:
    alpha = 0.00223
    lambd = 3379.7
    kelvin = temperature + 273.15

    if medium == "0.49% MC-PBS":
        a = 2.30e-6  # previously 2.23e-6, changed in Reichel2023 rev 2
        beta = -0.0056
    elif medium == "0.59% MC-PBS":
        a = 5.70e-6
        beta = -0.0744
    elif medium == "0.83% MC-PBS":
        a = 16.52e-6
        beta = -0.1455
    else:
        raise NotImplementedError(
            f"Medium {medium} not supported for model `buyukurganci-2022`!")

    k = a * np.exp(lambd / kelvin)
    n = alpha * kelvin + beta
    shear_rate = shear_rate_square_channel(flow_rate, channel_width, n)
    return k * shear_rate**(n - 1) * 1e3


def get_viscosity_mc_pbs_herold_2017(
        medium: Literal["0.49% MC-PBS", "0.59% MC-PBS"] = "0.49% MC-PBS",
        channel_width: float = 20.0,
        flow_rate: float = 0.16,
        temperature: float = 23.0):
    r"""Compute viscosity of MC-PBS according to :cite:`Herold2017`

    Note that all the factors in equation 5.2 in :cite:`Herold2017`
    compute to 8, which is essentially what is implemented in
    :func:`shear_rate_square_channel`:

    .. math::

        1.1856 \cdot 6 \cdot \frac{2}{3} \cdot \frac{1}{0.5928} = 8
    """
    # see figure (9) in Herold arXiv:1704.00572 (2017)
    check_temperature("'herold-2017' MC-PBS", temperature, 18, 26)
    # convert flow_rate from µL/s to m³/s
    # convert channel_width from µm to m
    term1 = 1.1856 * 6 * flow_rate * 1e-9 / (channel_width * 1e-6)**3 * 2 / 3

    if medium == "0.49% MC-PBS":
        temp_corr = (temperature / 23.2)**-0.866
        term2 = 0.6771 / 0.5928 + 0.2121 / (0.5928 * 0.677)
        eta = 0.179 * (term1 * term2)**(0.677 - 1) * temp_corr * 1e3
    elif medium == "0.59% MC-PBS":
        temp_corr = (temperature / 23.6)**-0.866
        term2 = 0.6771 / 0.5928 + 0.2121 / (0.5928 * 0.634)
        eta = 0.360 * (term1 * term2)**(0.634 - 1) * temp_corr * 1e3
    else:
        raise NotImplementedError(
            f"Medium {medium} not supported for model `herold-2017`!")
    return eta


def get_viscosity_water_kestin_1978(temperature: float = 23.0):
    """Compute the viscosity of water according to :cite:`Kestin_1978`"""
    # see equation (15) in Kestin et al, J. Phys. Chem. 7(3) 1978
    check_temperature("'kestin-1978' water", temperature, 0, 40)
    eta0 = 1.002  # [mPa s]
    right = (20-temperature) / (temperature + 96) \
        * (+ 1.2364
           - 1.37e-3 * (20 - temperature)
           + 5.7e-6 * (20 - temperature)**2
           )
    eta = eta0 * 10**right
    return eta



import re

from ..feat_basin import Basin

from .api import REQUESTS_AVAILABLE, APIHandler, DCORAccessError
from .base import RTDC_DCOR


REGEXP_FULL_DCOR_URL = re.compile(
    r"^https?:\/\/"  # scheme
    r"[a-z0-9-\.]*\.[a-z0-9-\.]*\/?api\/3\/action\/dcserv\?id="  # host and API
    r"[a-f0-9]{8}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{12}$")  # id


class DCORBasin(Basin):
    basin_format = "dcor"
    basin_type = "remote"

    def __init__(self, *args, **kwargs):
        """Access to private and public DCOR resources

        Since version 2 of the DCOR data API, all feature data are
        accessed via :class:`.HTTPBasin`s on S3. The DCOR basin is just
        a wrapper around those `HTTPBasin`s.

        For private resources, the DCOR format facilitates authentication
        via access tokens. Behind the scenes, DCOR creates a pre-signed
        URL to access private data on an S3 object storage provider.
        Note that you must let dclab know your DCOR access
        token via :func:`.APIHandler.add_api_key` for this to work.

        The `location` must be a full DCOR URL, including the scheme
        and netloc, e.g:

            https://dcor.mpl.mpg.de/api/3/action/dcserv?
            id=b1404eb5-f661-4920-be79-5ff4e85915d5
        """
        self._available_verified = None
        super(DCORBasin, self).__init__(*args, **kwargs)

    def _load_dataset(self, location, **kwargs):
        return RTDC_DCOR(location, **kwargs)

    def is_available(self):
        """Check whether a DCOR resource is available

        Notes
        -----
        - Make sure that your DCOR access token is stored in
          :class:`.APIHandler`. You can add tokens with
          :func:`.APIHandler.add_api_key`.
        """
        with self._av_check_lock:
            if not REQUESTS_AVAILABLE:
                # don't even bother
                self._available_verified = False
            elif not is_full_dcor_url(self.location):
                # not a full DCOR URL
                self._available_verified = False
            if self._available_verified is None:
                api = APIHandler(self.location)
                try:
                    self._available_verified = api.get("valid")
                except DCORAccessError:
                    self._available_verified = False
        return self._available_verified


def is_full_dcor_url(string):
    if not isinstance(string, str):
        return False
    else:
        return REGEXP_FULL_DCOR_URL.match(string.strip())



import numpy as np

from .ancillary_feature import AncillaryFeature


def compute_area_ratio(mm):
    valid = mm["area_msd"] != 0
    out = np.nan * np.ones(len(mm), dtype=float)
    return np.divide(mm["area_cvx"], mm["area_msd"], where=valid, out=out)


def compute_area_um(mm):
    pxs = mm.config["imaging"]["pixel size"]
    return mm["area_cvx"] * pxs**2


def compute_aspect(mm):
    """Compute the aspect ratio of the bounding box

    Notes
    -----
    If the cell is elongated along the channel, i.e.
    `size_x` is larger than `size_y`, then the aspect
    ratio is larger than 1.
    """
    out = np.nan * np.ones(len(mm), dtype=float)
    valid = mm["size_y"] != 0
    # parallel to flow, perpendicular to flow
    return np.divide(mm["size_x"], mm["size_y"], where=valid, out=out)


def compute_deform(mm):
    return 1 - mm["circ"]


def compute_index(mm):
    return np.arange(1, len(mm)+1)


def compute_time(mm):
    fr = mm.config["imaging"]["frame rate"]
    # Since version 0.47.8, we don't "normalize" the time anymore
    # with the information from mm["frame"][0]. This is important
    # for cases where it is important to know the time elapsed before
    # the first event was recorded (issue #207).
    return np.array(mm["frame"], dtype=float) / fr


AncillaryFeature(feature_name="time",
                 method=compute_time,
                 req_config=[["imaging", ["frame rate"]]],
                 req_features=["frame"])


AncillaryFeature(feature_name="index",
                 method=compute_index)


def register():
    AncillaryFeature(feature_name="area_ratio",
                     method=compute_area_ratio,
                     req_features=["area_cvx", "area_msd"])

    AncillaryFeature(feature_name="area_um",
                     method=compute_area_um,
                     req_config=[["imaging", ["pixel size"]]],
                     req_features=["area_cvx"])

    AncillaryFeature(feature_name="aspect",
                     method=compute_aspect,
                     req_features=["size_x", "size_y"])

    AncillaryFeature(feature_name="deform",
                     method=compute_deform,
                     req_features=["circ"])



"""Internal basins were introduced in version 0.60.0

They provide a few-to-many mapping for feature data and can be used
to achieve high compression ratios, an advantage that is primarily
of interest for the "image_bg" feature.
"""
import h5py
import numpy as np

from dclab import new_dataset, rtdc_dataset, RTDCWriter
from dclab.rtdc_dataset import feat_basin

import pytest

from helper_methods import retrieve_data


@pytest.mark.parametrize("mapping", [
    # monotonic
    np.array([0] * 2 + [1] * 3 + [2] * 1 + [3] * 4),
    # irregular access (useful for e.g. temperature)
    np.array([1, 0, 2, 3, 0, 2, 0, 1, 3, 1]),
])
def test_read_write_internal_basin_data(mapping):
    """write internal basin data and make sure everything is correct"""
    h5path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")
    h5path_small = h5path.with_name("smaller.rtdc")

    internal_data = {
        "userdef1": np.arange(4),
        "image_bg": np.concatenate([
            np.full((1, 80, 320), 145),
            np.full((1, 80, 320), 140),
            np.full((1, 80, 320), 150),
            np.full((1, 80, 320), 142),
            ])
    }

    # sanity checks
    with h5py.File(h5path) as h5:
        assert len(h5["events/deform"]) == 10
        assert len(h5["events/deform"]) == len(mapping)

    with h5py.File(h5path) as src, RTDCWriter(h5path_small) as hw:
        # first, copy all the scalar features to the new file
        rtdc_dataset.rtdc_copy(src_h5file=src,
                               dst_h5file=hw.h5file,
                               features="scalar")
        # second, define the internal basin
        hw.store_basin(basin_name="example basin",
                       basin_type="internal",
                       basin_format="h5dataset",
                       basin_locs=["basin_events"],
                       basin_descr="an example test basin",
                       internal_data=internal_data,
                       basin_map=mapping,
                       basin_feats=sorted(internal_data.keys()),
                       )

    # sanity checks
    with h5py.File(h5path_small) as h5:
        assert np.all(h5["basin_events/userdef1"][:]
                      == internal_data["userdef1"])
        assert np.all(h5["basin_events/image_bg"][:]
                      == internal_data["image_bg"])

    # now try to open the dataset containing the internal basin
    with new_dataset(h5path_small) as ds:
        for ii, mid in enumerate(mapping):
            assert np.all(ds["image_bg"][ii]
                          == internal_data["image_bg"][mid])
            assert np.all(ds["userdef1"][ii]
                          == internal_data["userdef1"][mid])


@pytest.mark.parametrize("mapping", [
    # monotonic
    np.array([0] * 2 + [1] * 3 + [2] * 1 + [3] * 4),
    # irregular access (useful for e.g. temperature)
    np.array([1, 0, 2, 3, 0, 2, 0, 1, 3, 1]),
])
def test_read_write_export_internal_basin_data_issue_262(mapping):
    """write internal basin data and make sure everything is correct"""
    h5path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")
    h5path_small = h5path.with_name("smaller.rtdc")
    h5path_export = h5path.with_name("exported.rtdc")

    internal_data = {
        "userdef1": np.arange(4),
        "image_bg": np.concatenate([
            np.full((1, 80, 320), 145),
            np.full((1, 80, 320), 140),
            np.full((1, 80, 320), 150),
            np.full((1, 80, 320), 142),
        ])
    }

    with h5py.File(h5path) as src, RTDCWriter(h5path_small) as hw:
        # first, copy all the scalar features to the new file
        rtdc_dataset.rtdc_copy(src_h5file=src,
                               dst_h5file=hw.h5file,
                               features="scalar")
        # second, define the internal basin
        hw.store_basin(basin_name="example basin",
                       basin_type="internal",
                       basin_format="h5dataset",
                       basin_locs=["basin_events"],
                       basin_descr="an example test basin",
                       internal_data=internal_data,
                       basin_map=mapping,
                       basin_feats=sorted(internal_data.keys()),
                       )

    # now try to open the dataset containing the internal basin
    # and export it to another .rtdc file.
    with new_dataset(h5path_small) as ds:
        ds.export.hdf5(
            path=h5path_export,
            features=["userdef1", "deform"],
            logs=True,
            tables=True,
            basins=True,  # this is tested in issue #262
            meta_prefix="",
            override=False)
        assert len(ds["deform"]) == 10
        assert len(ds["userdef1"][:]) == 10
        assert len(ds["userdef1"]) == 10

    with new_dataset(h5path_export) as dse:
        assert "image_bg" in dse.features_basin
        assert "image_bg" not in dse.features_innate
        assert np.all(dse["userdef1"] == internal_data["userdef1"][mapping])


def test_writer_error_internal_data():
    h5path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")
    h5path_small = h5path.with_name("smaller.rtdc")

    with h5py.File(h5path) as src, RTDCWriter(h5path_small) as hw:
        # first, copy all the scalar features to the new file
        rtdc_dataset.rtdc_copy(src_h5file=src,
                               dst_h5file=hw.h5file,
                               features="scalar")
        with pytest.raises(ValueError,
                           match="you must specify `internal_data`"):
            hw.store_basin(
                basin_name="example basin",
                basin_type="internal",
                basin_format="h5dataset",
                basin_locs=["basin_events"],
                basin_descr="an example test basin",
                basin_map=np.zeros(10),
                basin_feats=["userdef1"],
                )


def test_writer_error_basin_data():
    h5path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")
    h5path_small = h5path.with_name("smaller.rtdc")

    with h5py.File(h5path) as src, RTDCWriter(h5path_small) as hw:
        # first, copy all the scalar features to the new file
        rtdc_dataset.rtdc_copy(src_h5file=src,
                               dst_h5file=hw.h5file,
                               features="scalar")

        with pytest.raises(ValueError,
                           match="Could not find feature 'userdef1'"):
            hw.store_basin(
                basin_name="example basin",
                basin_type="internal",
                basin_format="h5dataset",
                basin_locs=["basin_events"],
                basin_descr="an example test basin",
                basin_map=np.zeros(10),
                basin_feats=["userdef1"],
                internal_data=hw.h5file.require_group("basin_events")
                )


def test_writer_error_feature_exists():
    h5path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")
    h5path_small = h5path.with_name("smaller.rtdc")

    with h5py.File(h5path) as src, RTDCWriter(h5path_small) as hw:
        # first, copy all the scalar features to the new file
        rtdc_dataset.rtdc_copy(src_h5file=src,
                               dst_h5file=hw.h5file,
                               features="scalar")
        igroup = hw.h5file.require_group("basin_events")
        hw.write_ndarray(igroup, "userdef1", np.arange(2))

        with pytest.raises(ValueError,
                           match="feature 'userdef1' already exists"):
            hw.store_basin(
                basin_name="example basin",
                basin_type="internal",
                basin_format="h5dataset",
                basin_locs=["basin_events"],
                basin_descr="an example test basin",
                basin_map=np.zeros(10),
                basin_feats=["userdef1"],
                internal_data={"userdef1": np.arange(2)},
                )


@pytest.mark.parametrize("h5_obj_delete", [
    # delete the entire basin_events directory
    "basin_events",
    # delete only the feature data
    "basin_events/userdef1"
])
def test_write_read_internal_basin_data_missing_issue_268(h5_obj_delete):
    """dclab should not fail when the internal data are missing"""
    h5path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")
    h5path_small = h5path.with_name("smaller.rtdc")
    mapping = np.array([1, 0, 2, 3, 0, 2, 0, 1, 3, 1])

    internal_data = {
        "userdef1": np.arange(4)
    }

    # sanity checks
    with h5py.File(h5path) as h5:
        assert len(h5["events/deform"]) == 10
        assert len(h5["events/deform"]) == len(mapping)

    with h5py.File(h5path) as src, RTDCWriter(h5path_small) as hw:
        # first, copy all the scalar features to the new file
        rtdc_dataset.rtdc_copy(src_h5file=src,
                               dst_h5file=hw.h5file,
                               features="scalar")
        # second, define the internal basin
        hw.store_basin(basin_name="example basin",
                       basin_type="internal",
                       basin_format="h5dataset",
                       basin_locs=["basin_events"],
                       basin_descr="an example test basin",
                       internal_data=internal_data,
                       basin_map=mapping,
                       basin_feats=sorted(internal_data.keys()),
                       )

    # sanity check
    with h5py.File(h5path_small) as h5:
        assert np.all(h5["basin_events/userdef1"][:]
                      == internal_data["userdef1"])

    # remove the basin data from the output file
    with h5py.File(h5path_small, "a") as h5:
        del h5[h5_obj_delete]

    with pytest.warns(feat_basin.BasinFeatureMissingWarning,
                      match="userdef1"):
        # now try to open the dataset containing the internal basin
        with new_dataset(h5path_small) as ds:
            assert "userdef1" not in ds


def test_write_read_internal_basin_data_missing_partial_issue_268():
    """dclab should not fail when the internal data are missing"""
    h5path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")
    h5path_small = h5path.with_name("smaller.rtdc")
    mapping = np.array([1, 0, 2, 3, 0, 2, 0, 1, 3, 1])

    internal_data = {
        "userdef1": np.arange(4),
        "userdef2": np.arange(4)
    }

    # sanity checks
    with h5py.File(h5path) as h5:
        assert len(h5["events/deform"]) == 10
        assert len(h5["events/deform"]) == len(mapping)

    with h5py.File(h5path) as src, RTDCWriter(h5path_small) as hw:
        # first, copy all the scalar features to the new file
        rtdc_dataset.rtdc_copy(src_h5file=src,
                               dst_h5file=hw.h5file,
                               features="scalar")
        # second, define the internal basin
        hw.store_basin(basin_name="example basin",
                       basin_type="internal",
                       basin_format="h5dataset",
                       basin_locs=["basin_events"],
                       basin_descr="an example test basin",
                       internal_data=internal_data,
                       basin_map=mapping,
                       basin_feats=sorted(internal_data.keys()),
                       )

    # sanity check
    with h5py.File(h5path_small) as h5:
        assert np.all(h5["basin_events/userdef1"][:]
                      == internal_data["userdef1"])

    # remove the basin data from the output file
    with h5py.File(h5path_small, "a") as h5:
        # Only delete one feature. The other one should still be there.
        del h5["basin_events/userdef1"]

    # now try to open the dataset containing the internal basin
    with pytest.warns(feat_basin.BasinFeatureMissingWarning,
                      match="userdef1"):
        with new_dataset(h5path_small) as ds:
            assert "userdef1" not in ds
            assert "userdef2" in ds
            assert "userdef2" in ds.features_basin


def test_writer_warning_uncommon_location():
    h5path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")
    h5path_small = h5path.with_name("smaller.rtdc")

    with h5py.File(h5path) as src, RTDCWriter(h5path_small) as hw:
        # first, copy all the scalar features to the new file
        rtdc_dataset.rtdc_copy(src_h5file=src,
                               dst_h5file=hw.h5file,
                               features="scalar")
        internal_data = hw.h5file.require_group("uncommon_loc")
        hw.write_ndarray(internal_data,
                         name="userdef1",
                         data=np.arange(2),
                         )
        with pytest.warns(UserWarning,
                          match="You specified an uncommon location"):
            hw.store_basin(
                basin_name="example basin",
                basin_type="internal",
                basin_format="h5dataset",
                basin_locs=["uncommon_loc"],
                basin_descr="an example test basin",
                basin_map=np.zeros(10),
                basin_feats=["userdef1"],
                internal_data=internal_data,
                )




import numpy as np
import dclab


def test_bin_width_doane():
    a = np.arange(100)
    b = dclab.kde_methods.bin_width_doane(a)
    assert np.allclose(b, 12.951578044133464)


def test_bin_width_percentile():
    a = np.arange(100)
    b = dclab.kde_methods.bin_width_percentile(a)
    assert np.allclose(b, 3.4434782608695653)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



"""Viscosity models for Young's modulus estimation

This example visualizes the different viscosity models for the
MC-PBS media implemented in dclab. We reproduce the lower left part of
figure 3 in :cite:`Reichel2023` (channel width is 20 µm).

"""
import matplotlib.pylab as plt
import matplotlib.lines as mlines
from matplotlib import cm
import numpy as np

from dclab.features.emodulus import viscosity


visc_res = {}

for medium in ["0.49% MC-PBS", "0.59% MC-PBS"]:
    visc_her = {}
    visc_buy = {}

    kwargs = {
        "medium": medium,
        "channel_width": 20.0,
        "temperature": np.linspace(19, 37, 100, endpoint=True),
    }

    flow_rate = np.arange(0.02, 0.13, 0.02)

    for fr in flow_rate:
        visc_her[fr] = viscosity.get_viscosity_mc_pbs_herold_2017(
            flow_rate=fr, **kwargs)
        visc_buy[fr] = viscosity.get_viscosity_mc_pbs_buyukurganci_2022(
            flow_rate=fr, **kwargs)

    visc_res[medium] = [visc_her, visc_buy]


fig, axes = plt.subplots(1, 2, figsize=(8, 4), sharey="all", sharex="all")
colors = [cm.get_cmap("viridis")(x) for x in np.linspace(.8, 0,
                                                         len(flow_rate))]

for ii, medium in enumerate(visc_res):
    visc_her, visc_buy = visc_res[medium]
    ax = axes.flatten()[ii]
    ax.set_title(medium)

    for jj, fr in enumerate(flow_rate):
        ax.plot(kwargs["temperature"], visc_her[fr], color=colors[jj], ls="--")
        ax.plot(kwargs["temperature"], visc_buy[fr], color=colors[jj], ls="-")

    ax.set_xlabel("Temperature [°C]")
    ax.set_ylabel("Viscosity [mPa·s]")
    ax.grid()
    ax.set_ylim(2, 12)

handles = []
for jj, fr in enumerate(flow_rate):
    handles.append(
        mlines.Line2D([], [], color=colors[jj], label=f'{fr:.4g} µL/s'))
handles.append(
    mlines.Line2D([], [], color='gray', label='Büyükurgancı 2022'))
handles.append(
    mlines.Line2D([], [], color='gray', ls="--", label='Herold 2017'))
axes[0].legend(handles=handles)


plt.tight_layout()
plt.show()



import json

import h5py
import numpy as np

import pytest

import dclab
from dclab import new_dataset, rtdc_dataset, RTDCWriter
from dclab.rtdc_dataset import feat_basin, fmt_http


from helper_methods import DCOR_AVAILABLE, retrieve_data


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.core.FeatureShouldExistButNotFoundWarning")
def test_basin_cyclic_dependency_found():
    """A basin can be defined in one of its sub-basins

    This is something dclab identifies and then
    raises a CyclicBasinDependencyFoundWarning
    """
    h5path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")
    paths_list = [
        h5path.with_name("level1.rtdc"),
        h5path.with_name("level2.rtdc"),
        h5path.with_name("level3.rtdc"),
    ]

    paths_list_rolled = np.roll(paths_list, 1)
    basins_list = []
    for pp in paths_list_rolled:
        basins_list.append({
            "basin_name": "user-defined basin",
            "basin_type": "file",
            "basin_format": "hdf5",
            "basin_feats": ["userdef1"],
            "basin_locs": [str(pp)],
            "verify": False,
        })

    for bdict, pp in zip(basins_list, paths_list):
        with h5py.File(h5path) as src, RTDCWriter(pp) as hw:
            # copy all the scalar features to the new file
            rtdc_dataset.rtdc_copy(src_h5file=src,
                                   dst_h5file=hw.h5file,
                                   features="scalar")
            hw.store_basin(**bdict)

    # Open the dataset
    with dclab.new_dataset(paths_list[0]) as ds:
        assert np.allclose(ds["deform"][0], 0.02494624), "sanity check"
        with pytest.warns(feat_basin.CyclicBasinDependencyFoundWarning,
                          match="Encountered cyclic basin dependency"):
            with pytest.raises(KeyError):
                _ = ds["userdef1"]


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.core.FeatureShouldExistButNotFoundWarning")
def test_basin_cyclic_dependency_found_2():
    """A basin can be defined in one of its sub-basins

    This is something dclab identifies and then
    raises a CyclicBasinDependencyFoundWarning
    """
    h5path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")
    # Those are the files with a cyclic dependency
    p1 = h5path.with_name("level1.rtdc")
    p2 = h5path.with_name("level2.rtdc")
    p3 = h5path.with_name("level3.rtdc")
    # This is the file that will contain the userdef1 feature
    pz = h5path.with_name("final.rtdc")

    # Initialize datasets
    for pp in [p1, p2, p3, pz]:
        with h5py.File(h5path) as src, RTDCWriter(pp) as hw:
            # copy all the scalar features to the new file
            rtdc_dataset.rtdc_copy(src_h5file=src,
                                   dst_h5file=hw.h5file,
                                   features="scalar")

    with RTDCWriter(pz) as hw:
        hw.store_feature("userdef1", hw.h5file["events/deform"][:])

    bn_kwargs = {
            "basin_type": "file",
            "basin_format": "hdf5",
            "basin_feats": ["userdef1"],
            "verify": False,
        }

    with RTDCWriter(p1) as hw:
        hw.store_basin(
            basin_name="link to path 2",
            basin_locs=[str(p2)],
            **bn_kwargs
        )

        # store a basin with a key that is sorted after the basin above
        bdat = {
            "name": "user-defined data",
            "type": "file",
            "format": "hdf5",
            "features": ["userdef1"],
            "paths": [str(pz)]
        }
        blines = json.dumps(bdat, indent=2).split("\n")
        basins = hw.h5file.require_group("basins")
        hw.write_text(basins, "zzzz99999zzzzz99999", blines)

    with RTDCWriter(p2) as hw:
        hw.store_basin(
            basin_name="link to path 3",
            basin_locs=[str(p3)],
            **bn_kwargs
        )

    with RTDCWriter(p3) as hw:
        hw.store_basin(
            basin_name="link to path 1, completing the circle",
            basin_locs=[str(p1)],
            **bn_kwargs
        )

    # Open the dataset
    with dclab.new_dataset(p1) as ds:
        assert np.allclose(ds["deform"][0], 0.02494624), "sanity check"
        assert "userdef1" in ds
        assert ds.basins[0].name == "link to path 2", "order matters"
        assert ds.basins[1].name == "user-defined data", "order matters"
        with pytest.warns(feat_basin.CyclicBasinDependencyFoundWarning,
                          match="Encountered cyclic basin dependency"):
            _ = ds["userdef1"]


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.filterwarnings(
    "ignore::UserWarning")
def test_basin_feature_image_export_from_basin_with_hierarchy(tmp_path):
    h5path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")
    epath = tmp_path / "exported.rtdc"
    epath2 = tmp_path / "exported2.rtdc"
    # Create a basins-only dataset using the export functionality
    with dclab.new_dataset(h5path) as ds:
        assert "image" in ds
        ds.export.hdf5(path=epath,
                       features=[],
                       basins=True,
                       filtered=False,
                       )
    # Attempt to export image data from the exported dataset
    with dclab.new_dataset(epath) as ds:
        ds2 = dclab.new_dataset(ds)
        assert "image" in ds2
        assert "image" not in ds2.features_innate
        ds2.export.hdf5(path=epath2,
                        features=["mask", "image"],
                        basins=False,
                        filtered=False,
                        )

    # Open the exported dataset and attempt to access the image feature
    with dclab.new_dataset(epath2) as ds:
        assert "mask" in ds
        assert np.any(ds["mask"][0])


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.skipif(not DCOR_AVAILABLE, reason="DCOR is not available")
@pytest.mark.parametrize("filtered", [True, False])
def test_basin_feature_nested_image_exported_available(tmp_path, filtered):
    pytest.importorskip("requests")
    epath = tmp_path / "exported.rtdc"
    # Load data from DCOR and export basin-only dataset
    with dclab.new_dataset("fb719fb2-bd9f-817a-7d70-f4002af916f0") as ds:
        assert "image" in ds
        ds.export.hdf5(path=epath,
                       features=[],
                       basins=True,
                       filtered=filtered,
                       )
    # Open the exported dataset and attempt to access the image feature
    with dclab.new_dataset(epath) as ds:
        assert "image" in ds
        assert np.any(ds["image"][0])
        imcount = 0
        for bn in ds.basins:
            if "image" in bn.ds:
                assert np.any(bn.ds["image"][0])
                imcount += 1
        assert imcount > 0, "image should be *some*-where"


def test_basin_hierarchy_trace():
    h5path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")
    h5path_small = h5path.with_name("smaller.rtdc")

    # Dataset creation
    with h5py.File(h5path) as src, RTDCWriter(h5path_small) as hw:
        # first, copy all the scalar features to the new file
        rtdc_dataset.rtdc_copy(src_h5file=src,
                               dst_h5file=hw.h5file,
                               features="scalar")
        # Next, store the basin information in the new dataset
        bdat = {
            "type": "file",
            "format": "hdf5",
            "features": ["trace"],
            "paths": [
                str(h5path),  # absolute path name
            ]
        }
        blines = json.dumps(bdat, indent=2).split("\n")
        basins = hw.h5file.require_group("basins")
        hw.write_text(basins, "trace_basin", blines)

    ds = dclab.new_dataset(h5path_small)
    assert "trace" not in ds.features_innate
    ds2 = dclab.new_dataset(ds)
    assert "trace" in ds
    assert "trace" in ds2
    assert "fl1_raw" in ds["trace"]
    assert np.allclose(
        np.mean(ds2["trace"]["fl1_raw"][0]),
        24.5785536159601,
        atol=0, rtol=1e-5
    )


def test_basin_hierarchy_trace_missing():
    h5path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")
    h5path_small = h5path.with_name("smaller.rtdc")

    # Dataset creation
    with h5py.File(h5path) as src, RTDCWriter(h5path_small) as hw:
        # first, copy all the scalar features to the new file
        rtdc_dataset.rtdc_copy(src_h5file=src,
                               dst_h5file=hw.h5file,
                               features="scalar")
        # Next, store the basin information in the new dataset
        bdat = {
            "type": "file",
            "format": "hdf5",
            "features": ["trace"],
            "paths": [
                str(h5path),  # absolute path name
            ]
        }
        blines = json.dumps(bdat, indent=2).split("\n")
        basins = hw.h5file.require_group("basins")
        hw.write_text(basins, "trace missing", blines)

    h5path.unlink()

    ds = dclab.new_dataset(h5path_small)
    ds2 = dclab.new_dataset(ds)
    ds2.apply_filter()
    assert "trace" not in ds2
    with pytest.raises(KeyError, match="does not contain the feature"):
        ds2["trace"]


@pytest.mark.skipif(not DCOR_AVAILABLE, reason="DCOR is not available")
def test_basin_not_allowed_to_have_local_basin_in_remote_basin():
    """Since version 0.57.5 we do forbid remote datasets with local basins

    This type of basin-inception would allow an attacker to access files
    on the local file system if the instance was passed on to him via a
    script or service. In addition, defining a local basin in a remote
    dataset does not have a sensible use case.
    """
    pytest.importorskip("requests")
    # sanity check
    s3_url = ("https://objectstore.hpccloud.mpcdf.mpg.de/"
              "circle-5a7a053d-55fb-4f99-960c-f478d0bd418f/"
              "resource/fb7/19f/b2-bd9f-817a-7d70-f4002af916f0")
    with fmt_http.RTDC_HTTP(s3_url) as dss3:
        assert not dss3._local_basins_allowed, "sanity check"

    # Because wo do not have the resources to upload anything and access
    # it via HTTP to test this, we simply modify the `_local_basins_allowed`
    # property of an `RTDC_HDF5` instance. This is a hacky workaround, but
    # it should be fine.

    h5path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")
    h5path_small = h5path.with_name("smaller.rtdc")

    # Dataset creation
    with h5py.File(h5path) as src, RTDCWriter(h5path_small) as hw:
        # first, copy all the scalar features to the new file
        rtdc_dataset.rtdc_copy(src_h5file=src,
                               dst_h5file=hw.h5file,
                               features="scalar")
        # Next, store the basin information in the new dataset
        bdat = {
            "type": "file",
            "format": "hdf5",
            "features": ["image"],
            "paths": [
                str(h5path),  # absolute path name
            ]
        }
        blines = json.dumps(bdat, indent=2).split("\n")
        basins = hw.h5file.require_group("basins")
        hw.write_text(basins, "image_based_local_basin", blines)

    # sanity check
    with dclab.new_dataset(h5path_small) as ds:
        assert ds._local_basins_allowed, "sanity check"
        assert "image" in ds

    with dclab.new_dataset(h5path_small) as ds2:
        ds2._local_basins_allowed = False
        with pytest.warns(
            UserWarning,
                match="Basin type 'file' not allowed for format 'hdf5"):
            # Read the comments above carefully, if you wonder why this is.
            assert "image" not in ds2, "not there, local basins are forbidden"


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.parametrize("path_sep", [r"/", r"\\"])
def test_basin_relative_paths(path_sep):
    """Relative paths should work for both `\\` and ¸`/`."""
    h5path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")
    h5path_small = h5path.parent / "subdirectory" / "relative.rtdc"
    h5path_small.parent.mkdir()

    with dclab.new_dataset(h5path) as ds0:
        assert "image" in ds0

    # Dataset creation
    with h5py.File(h5path) as src, RTDCWriter(h5path_small) as hw:
        # first, copy all the scalar features to the new file
        rtdc_dataset.rtdc_copy(src_h5file=src,
                               dst_h5file=hw.h5file,
                               features="scalar")
        # Next, store the basin information in the new dataset
        bdat = {
            "type": "file",
            "format": "hdf5",
            "features": ["image"],
            "paths": [f"..{path_sep}{h5path.name}"]
        }
        blines = json.dumps(bdat, indent=2).split("\n")
        basins = hw.h5file.require_group("basins")
        hw.write_text(basins, "relative_paths_basin", blines)

    with dclab.new_dataset(h5path_small) as ds:
        assert "image" not in ds.features_innate
        assert "image" in ds


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_basin_key_reproducible():
    """Since 0.62.9 we sort basin keys"""
    h5path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")
    h5path_small = h5path.parent / "subdirectory" / "relative.rtdc"
    h5path_small.parent.mkdir()
    h5path_small_2 = h5path_small.with_name("from_writer.rtdc")

    # Dataset creation
    with h5py.File(h5path) as src, RTDCWriter(h5path_small) as hw:
        # first, copy all the scalar features to the new file
        rtdc_dataset.rtdc_copy(src_h5file=src,
                               dst_h5file=hw.h5file,
                               features="scalar")
        # Next, store the basin information in the new dataset manually
        bdat = {
            "type": "file",
            "format": "hdf5",
            "features": ["image"],
            "paths": [f"../{h5path.name}"]
        }
        blines = json.dumps(bdat, indent=2).split("\n")
        basins = hw.h5file.require_group("basins")
        hw.write_text(basins, "relative_paths_basin", blines)

    # When loading a basin, dclab should use the existing key
    # from the input file.
    with dclab.new_dataset(h5path_small) as ds:
        assert len(ds.basins) == 1
        assert ds.basins[0].key == "relative_paths_basin"

    # By default, dclab should store the basin with a key made from
    # its hash.
    # Dataset creation
    with h5py.File(h5path) as src, RTDCWriter(h5path_small_2) as hw:
        # first, copy all the scalar features to the new file
        rtdc_dataset.rtdc_copy(src_h5file=src,
                               dst_h5file=hw.h5file,
                               features="scalar")
        # Next, store the basin information in the new dataset with dclab
        hw.store_basin(
            basin_name="relative_paths_basin",
            basin_type="file",
            basin_format="hdf5",
            basin_feats=["image"],
            basin_locs=[f"../{h5path.name}"],
            verify=False,
        )
    # When loading a basin, dclab should use the existing key
    # from the input file.
    with dclab.new_dataset(h5path_small_2) as ds:
        assert len(ds.basins) == 1
        # dclab 0.62.9
        assert ds.basins[0].key in [
            # posix linesep
            "b8763fad05e17b58407875fe9c28d90b",
            # win linesep
            "a495fd2d495e0d991f3a4a313ec3a3fc",
        ]


def test_basin_sorting_basic():
    bnlist = [
        {"type": "remote", "format": "dcor", "ident": 0},
        {"type": "file", "format": "hdf5", "ident": 1},
        {"type": "hans", "format": "hdf5", "ident": 2},
        {"type": "remote", "format": "http", "ident": 3},
    ]
    sorted_list = sorted(bnlist, key=feat_basin.basin_priority_sorted_key)
    assert sorted_list[0]["ident"] == 1
    assert sorted_list[1]["ident"] == 3
    assert sorted_list[2]["ident"] == 0
    assert sorted_list[3]["ident"] == 2


@pytest.mark.parametrize("btype,bformat,sortval", [
    ["internal", "h5dataset", "aaa"],
    ["file", "hdf5", "bba"],
    ["remote", "http", "cca"],
    ["remote", "s3", "cda"],
    ["remote", "dcor", "cea"],
    ["peter", "hdf5", "zba"],
    ["remote", "hans", "cza"],
    ["hans", "peter", "zza"],
]
                         )
def test_basin_sorting_key(btype, bformat, sortval):
    bdict = {"type": btype,
             "format": bformat,
             }
    assert feat_basin.basin_priority_sorted_key(bdict) == sortval


def test_basin_unsupported_basin_format():
    h5path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")
    h5path_small = h5path.with_name("smaller.rtdc")

    # Dataset creation
    with h5py.File(h5path) as src, RTDCWriter(h5path_small) as hw:
        # first, copy all the scalar features to the new file
        rtdc_dataset.rtdc_copy(src_h5file=src,
                               dst_h5file=hw.h5file,
                               features="scalar")
        # Next, store the basin information in the new dataset
        bdat = {
            "type": "file",
            "format": "peter",
            "paths": [
                "fake.rtdc",  # fake path
                str(h5path),  # absolute path name
            ]
        }
        blines = json.dumps(bdat, indent=2).split("\n")
        basins = hw.h5file.require_group("basins")
        hw.write_text(basins, "invalid_format_basin", blines)

    h5path.unlink()

    with pytest.warns(UserWarning,
                      match="Encountered unsupported basin format 'peter'"):
        with new_dataset(h5path_small) as ds:
            assert "image" not in ds
            assert not ds.features_basin


def test_basin_unsupported_basin_type():
    h5path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")
    h5path_small = h5path.with_name("smaller.rtdc")

    # Dataset creation
    with h5py.File(h5path) as src, RTDCWriter(h5path_small) as hw:
        # first, copy all the scalar features to the new file
        rtdc_dataset.rtdc_copy(src_h5file=src,
                               dst_h5file=hw.h5file,
                               features="scalar")
        # Next, store the basin information in the new dataset
        bdat = {
            "type": "peter",
            "format": "hdf5",
            "paths": [
                "fake.rtdc",  # fake path
                str(h5path),  # absolute path name
            ]
        }
        blines = json.dumps(bdat, indent=2).split("\n")
        basins = hw.h5file.require_group("basins")
        hw.write_text(basins, "invalid_type_basin", blines)

    h5path.unlink()

    with pytest.warns(UserWarning,
                      match="Encountered unsupported basin type 'peter'"):
        with new_dataset(h5path_small) as ds:
            assert "image" not in ds
            assert not ds.features_basin



from ..meta_table import MetaTable


class H5Tables:
    def __init__(self, h5):
        self.h5file = h5
        self._cache_keys = None

    def __getitem__(self, key):
        if key in self.keys():
            tab = H5Table(self.h5file["tables"][key])
        else:
            raise KeyError(f"Table '{key}' not found or empty "
                           f"in {self.h5file.file.filename}!")
        return tab

    def __iter__(self):
        # dict-like behavior
        for key in self.keys():
            yield key

    def __len__(self):
        return len(self.keys())

    def keys(self):
        if self._cache_keys is None:
            names = []
            if "tables" in self.h5file:
                for key in self.h5file["tables"]:
                    if self.h5file["tables"][key].size:
                        names.append(key)
            self._cache_keys = names
        return self._cache_keys


class H5Table(MetaTable):
    def __init__(self, h5_ds):
        self._h5_ds = h5_ds
        self._keys = None
        self._meta = None

    def __array__(self, *args, **kwargs):
        return self._h5_ds.__array__(*args, **kwargs)

    @property
    def meta(self):
        if self._meta is None:
            self._meta = dict(self._h5_ds.attrs)
        return self._meta

    def has_graphs(self):
        return self.keys() is not None

    def keys(self):
        if self._keys is None:
            self._keys = self._h5_ds.dtype.names
        return self._keys

    def __getitem__(self, key):
        return self._h5_ds[key]



import json
import uuid

import h5py
import numpy as np

import pytest

from dclab import new_dataset, RTDCWriter
from dclab.rtdc_dataset.fmt_dcor import DCORBasin, RTDC_DCOR
from dclab.rtdc_dataset.feat_basin import BasinNotAvailableError

from helper_methods import DCOR_AVAILABLE, retrieve_data


pytest.importorskip("requests")

if not DCOR_AVAILABLE:
    pytest.skip("No connection to DCOR", allow_module_level=True)

dcor_url = ("https://dcor.mpl.mpg.de/api/3/action/dcserv?id="
            "fb719fb2-bd9f-817a-7d70-f4002af916f0")


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_basin_as_dict(tmp_path):
    tmp_path = tmp_path.resolve()
    h5path = tmp_path / "test_basin_dcor.rtdc"

    with h5py.File(h5path, "a") as dst, RTDC_DCOR(dcor_url) as src:
        # Store non-existent basin information
        with RTDCWriter(dst, mode="append") as hw:
            meta = src.config.as_dict(pop_filtering=True)
            hw.store_metadata(meta)
            hw.store_basin(basin_name="example basin",
                           basin_type="remote",
                           basin_format="dcor",
                           basin_locs=[dcor_url],
                           basin_descr="an example DCOR test basin",
                           )

    with new_dataset(h5path) as ds:
        assert ds._enable_basins
        bdict = ds.basins[0].as_dict()
        assert bdict["basin_name"] == "example basin"
        assert bdict["basin_type"] == "remote"
        assert bdict["basin_format"] == "dcor"
        assert bdict["basin_locs"] == [dcor_url]
        assert bdict["basin_descr"] == "an example DCOR test basin"

    # Now use the data from `bdict` to create a new basin
    h5path_two = h5path.with_name("smaller_two.rtdc")

    # Dataset creation
    with RTDCWriter(h5path_two) as hw:
        # first, copy all the scalar features to the new file
        hw.store_metadata(meta)
        hw.store_basin(**bdict)

    with new_dataset(h5path_two) as ds2:
        bdict2 = ds2.basins[0].as_dict()
        assert bdict2["basin_name"] == "example basin"
        assert bdict2["basin_type"] == "remote"
        assert bdict2["basin_format"] == "dcor"
        assert bdict2["basin_locs"] == [dcor_url]
        assert bdict2["basin_descr"] == "an example DCOR test basin"


@pytest.mark.parametrize("url", [
    "https://example.com/nonexistentbucket/nonexistentkey",
    f"https://objectstore.hpccloud.mpcdf.mpg.de/noexist-{uuid.uuid4()}/key",
])
@pytest.mark.filterwarnings(
    "ignore::dclab.http_utils.ConnectionTimeoutWarning")
def test_basin_not_available(url):
    h5path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")

    # Dataset creation
    with h5py.File(h5path, "a") as dst:
        # Store non-existent basin information
        bdat = {
            "type": "remote",
            "format": "dcor",
            "urls": [
                # does not exist
                url
            ]
        }
        blines = json.dumps(bdat, indent=2).split("\n")
        basins = dst.require_group("basins")
        with RTDCWriter(dst, mode="append") as hw:
            hw.write_text(basins, "my_basin", blines)

    # Open the dataset and check whether basin is missing
    with new_dataset(h5path) as ds:
        assert not ds.features_basin
        # This is a very subtle test for checking whether invalid basins
        # are just ignored:
        _ = ds["index"]

    # Also test that on a lower level
    bn = DCORBasin("https://dcor.mpl.mpg.de/api/3/action/dcserv?id="
                   "00000000-0000-0000-0000-000000000000")
    assert not bn.is_available()
    with pytest.raises(BasinNotAvailableError, match="is not available"):
        _ = bn.ds


def test_create_basin_file_non_matching_identifier(tmp_path):
    h5path = tmp_path / "test_basin_dcor.rtdc"

    with h5py.File(h5path, "a") as dst, RTDC_DCOR(dcor_url) as src:
        # Store non-existent basin information
        bdat = {
            "type": "remote",
            "format": "dcor",
            "urls": [dcor_url],
            "features": ["deform"],
        }
        blines = json.dumps(bdat, indent=2).split("\n")
        basins = dst.require_group("basins")
        with RTDCWriter(dst, mode="append") as hw:
            hw.write_text(basins, "my_basin", blines)
            meta = src.config.as_dict(pop_filtering=True)
            meta["experiment"]["run identifier"] = "hoolahoop"
            hw.store_metadata(meta)

    with new_dataset(h5path) as ds:
        assert ds.basins
        # The feature shows up as available...
        assert ds.features_basin == ["deform"]
        # ...but it is actually not, since the run identifier does not match
        # and therefore dclab does not allow the user to access it.
        #
        # Until a workaround is found for invalid basin URLs that return a
        # status code of 200, do this test which should raise a warning,
        # because `__contains__` returns True for "trace", but the trace data
        # are nowhere to find.
        with (pytest.warns(UserWarning, match="but I cannot get its data"),
              pytest.raises(KeyError, match="deform")):
            _ = ds["deform"]


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_create_basin_file_with_no_data(tmp_path):
    h5path = tmp_path / "test_basin_dcor.rtdc"

    with h5py.File(h5path, "a") as dst, RTDC_DCOR(dcor_url) as src:
        # Store non-existent basin information
        bdat = {
            "type": "remote",
            "format": "dcor",
            "urls": [dcor_url]
        }
        blines = json.dumps(bdat, indent=2).split("\n")
        basins = dst.require_group("basins")
        with RTDCWriter(dst, mode="append") as hw:
            hw.write_text(basins, "my_basin", blines)
            meta = src.config.as_dict(pop_filtering=True)
            hw.store_metadata(meta)

    with new_dataset(h5path) as ds:
        # This is essentially a nested basin features test. The basin is
        # a DCOR dataset which has two basins, the condensed version of the
        # data and the full version of the data as HTTP basins.
        assert len(ds.basins) == 1
        bn = ds.basins[0]
        assert len(bn.ds.basins) == 2
        assert ds.features_basin
        assert len(ds) == 5000
        assert np.allclose(ds["deform"][0], 0.009741939,
                           atol=0, rtol=1e-5)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_create_basin_file_with_one_feature(tmp_path):
    h5path = tmp_path / "test_basin_dcor.rtdc"

    with h5py.File(h5path, "a") as dst, RTDC_DCOR(dcor_url) as src:
        # Store non-existent basin information
        bdat = {
            "type": "remote",
            "format": "dcor",
            "urls": [dcor_url],
            "features": ["deform"],
        }
        blines = json.dumps(bdat, indent=2).split("\n")
        basins = dst.require_group("basins")
        with RTDCWriter(dst, mode="append") as hw:
            hw.write_text(basins, "my_basin", blines)
            meta = src.config.as_dict(pop_filtering=True)
            hw.store_metadata(meta)

    with new_dataset(h5path) as ds:
        assert ds.features_basin
        assert len(ds) == 5000
        assert "deform" in ds.features_basin
        assert "area_um" not in ds.features_basin
        assert "deform" in ds
        assert "area_um" not in ds
        assert np.allclose(ds["deform"][0], 0.009741939,
                           atol=0, rtol=1e-5)



"""
.. versionadded:: 0.34.0
"""
from __future__ import annotations

import hashlib
import importlib
import pathlib
import sys
from typing import List, Optional

from ...definitions import feat_logic
from ..feat_anc_core import AncillaryFeature


class PluginImportError(BaseException):
    pass


class PlugInFeature(AncillaryFeature):
    def __init__(self,
                 feature_name: str,
                 info: dict,
                 plugin_path: Optional[str | pathlib.Path] = None):
        """A user-defined plugin feature

        Parameters
        ----------
        feature_name: str
            name of a feature that matches that defined in `info`
        info: dict
            Full plugin recipe (for all features) as given in the
            `info` dictionary in the plugin file.
            At least the following keys must be specified:

            - "method": callable function computing the plugin feature
              values (takes an :class`dclab.rtdc_dataset.core.RTDCBase`
              as argument)
            - "feature names": list of plugin feature names provided
              by the plugin

            The following features are optional:

            - "description": short (one-line) description of the plugin
            - "long description": long description of the plugin
            - "feature labels": feature labels used e.g. for plotting
            - "feature shapes": list of tuples for each feature
              indicating the shape (this is required only for
              non-scalar features; for scalar features simply set
              this to ``None`` or ``(1,)``).
            - "scalar feature": list of boolean values indicating
              whether the features are scalar
            - "config required": configuration keys required to
              compute the plugin features (see the `req_config`
              parameter for :class:`.AncillaryFeature`)
            - "features required": list of feature names required to
              compute the plugin features (see the `req_features`
              parameter for :class:`.AncillaryFeature`)
            - "method check required": additional method that checks
              whether the features can be computed (see the `req_func`
              parameter for :class:`.AncillaryFeature`)
            - "version": version of this plugin (please use
              semantic verioning)

        plugin_path: str or pathlib.Path, optional
            path which was used to load the `PlugInFeature` with
            :func:`load_plugin_feature`.

        Notes
        -----
        `PluginFeature` inherits from :class:`AncillaryFeature
        <dclab.rtdc_dataset.feat_anc_core.ancillary_feature.AncillaryFeature>`.
        Please read the advanced section on `PluginFeatures` in the dclab docs.
        """
        if plugin_path is not None:
            plugin_path = pathlib.Path(plugin_path)

        #: Plugin feature name
        self.feature_name = feature_name
        #: Path to the original plugin file
        self.plugin_path = plugin_path

        # perform sanity checks
        self._sanity_check_original_info(info)
        # keep this for tests
        self._original_info = info

        # populate `info` dictionary with missing values
        #: Dictionary containing all information relevant for
        #: this particular plugin feature instance
        self.plugin_feature_info = self._process_plugin_info(info)

        # register this plugin feature in definitions
        # This must be executed before initializing the super class
        # (AncillaryFeature). If we don't do this, then `remove_plugin_feature`
        # may end up removing innate features e.g., "deform".
        feat_logic.feature_register(
            name=self.feature_name,
            label=self.plugin_feature_info["feature label"],
            is_scalar=self.plugin_feature_info["scalar feature"]
        )

        # Instantiate the super class
        super(PlugInFeature, self).__init__(
            feature_name=self.plugin_feature_info["feature name"],
            method=self.plugin_feature_info["method"],
            req_config=self.plugin_feature_info["config required"],
            req_features=self.plugin_feature_info["features required"],
            req_func=self.plugin_feature_info["method check required"],
            identifier=self.plugin_feature_info["identifier"],
        )

    def _process_plugin_info(self, original_info: dict) -> dict:
        """Return dictionary with all relevant info for this instance
        """
        fidx = original_info["feature names"].index(self.feature_name)

        # determine feature label
        if ("feature labels" in original_info
                and original_info["feature labels"][fidx]):
            label = original_info["feature labels"][fidx]
        else:
            label = f"Plugin feature {self.feature_name}"

        # determine whether we have a scalar feature
        if "scalar feature" in original_info:
            is_scalar = original_info["scalar feature"][fidx]
        else:
            is_scalar = True  # default

        if is_scalar:
            event_shape = (1,)
        else:
            if "feature shapes" in original_info:
                event_shape = original_info["feature shapes"][fidx]
            else:
                event_shape = None

        # We assume that the script does not import any other custom
        # Python scripts.
        md5hasher = hashlib.md5()
        if isinstance(self.plugin_path, pathlib.Path):
            md5hasher.update(self.plugin_path.read_bytes())
        else:
            md5hasher.update(original_info["method"].__code__.co_code)
        md5hasher.update(self.feature_name.encode("utf-8"))
        md5hasher.update(original_info.get("version", "").encode("utf-8"))
        for feat in original_info.get("features required", []):
            md5hasher.update(feat.encode("utf-8"))
        identifier = md5hasher.hexdigest()

        feature_info = {
            "method": original_info["method"],
            "description": original_info.get(
                "description", "No description provided"),
            "long description": original_info.get(
                "long description", "No long description provided."),
            "feature name": self.feature_name,
            "feature label": label,
            "feature shape": event_shape,
            "features required": original_info.get("features required", []),
            "config required": original_info.get("config required", []),
            "method check required": original_info.get(
                "method check required", lambda x: True),
            "scalar feature": is_scalar,
            # allow comparisons with distutil.version.LooseVersion
            "version": original_info.get("version", "0.0.0-unknown"),
            "plugin path": self.plugin_path,
            "identifier": identifier,
        }

        return feature_info

    def _sanity_check_original_info(self, original_info: dict):
        """Various checks on the `original_info` attribute dict

        Raises
        ------
        ValueError
            If the parameter `original_info` is not a dict.
            If the `self.feature_name` is not in
                `original_info["feature names"]`.
            If the `method` provided in parameter `original_info`
                is not callable.
        """
        if not isinstance(original_info, dict):
            raise ValueError(
                "PlugInFeature parameter for `info` must be a dict, instead "
                f"a '{type(original_info)}' was given.")

        if not isinstance(original_info["feature names"], list):
            raise ValueError("'feature names' must be a list, "
                             f"got '{type(original_info['feature names'])}'")

        if self.feature_name not in original_info["feature names"]:
            raise ValueError(
                f"The feature name '{self.feature_name}' is not defined in "
                + "the `info` dict of the plugin feature"
                + (f" at {self.plugin_path}" if self.plugin_path else "")
                + f". Defined names are '{original_info['feature names']}'.")

        if not callable(original_info["method"]):
            raise ValueError(
                "The `method` you have provided in the parameter `info` is "
                f"not callable ('{original_info['method']}' is not "
                "a function).")


def import_plugin_feature_script(
        plugin_path: str | pathlib.Path) -> dict:
    """Import the user-defined recipe and return the info dictionary

    Parameters
    ----------
    plugin_path: str or Path
        pathname to a valid dclab plugin script

    Returns
    -------
    info: dict
        Dictionary with the information required to instantiate
        one (or multiple) :class:`PlugInFeature`.

    Raises
    ------
    PluginImportError
        If the plugin can not be found

    Notes
    -----
    One recipe may define multiple plugin features.
    """
    path = pathlib.Path(plugin_path)
    if not path.exists():
        raise PluginImportError("The plugin could be not be found at "
                                f"'{plugin_path}'!")
    try:
        # insert the plugin directory to sys.path so we can import it
        sys.path.insert(-1, str(path.parent))
        sys.dont_write_bytecode = True
        plugin = importlib.import_module(path.stem)
    except BaseException as e:
        raise PluginImportError(
            f"The plugin {plugin_path} could not be loaded!") from e
    finally:
        # undo our path insertion
        sys.path.pop(0)
        sys.dont_write_bytecode = False

    return plugin.info


def load_plugin_feature(
        plugin_path: str | pathlib.Path) -> List[PlugInFeature]:
    """Find and load PlugInFeature(s) from a user-defined recipe

    Parameters
    ----------
    plugin_path: str or Path
        pathname to a valid dclab plugin Python script

    Returns
    -------
    plugin_list: list of PlugInFeature
        list of PlugInFeature instances loaded from `plugin_path`

    Raises
    ------
    ValueError
        If the script dictionary "feature names" are not a list

    Notes
    -----
    One recipe may define multiple plugin features.

    See Also
    --------
    import_plugin_feature_script: function that imports the plugin script
    PlugInFeature: class handling the plugin feature information
    dclab.rtdc_dataset.feat_temp.register_temporary_feature: alternative
        method for creating user-defined features
    """
    info = import_plugin_feature_script(plugin_path)
    if not isinstance(info["feature names"], list):
        raise ValueError(
            "'feature names' must be a list of strings.")

    plugin_list = []
    for feature_name in info["feature names"]:
        plugin_list.append(PlugInFeature(feature_name, info, plugin_path))
    return plugin_list


def remove_all_plugin_features():
    """Convenience function for removing all `PlugInFeature` instances

    See Also
    --------
    remove_plugin_feature: remove a single `PlugInFeature` instance
    """
    for plugin_instance in reversed(PlugInFeature.features):
        if isinstance(plugin_instance, PlugInFeature):
            remove_plugin_feature(plugin_instance)


def remove_plugin_feature(plugin_instance: PlugInFeature):
    """Convenience function for removing a `PlugInFeature` instance

    Parameters
    ----------
    plugin_instance: PlugInFeature
        The `PlugInFeature` instance to be removed from dclab

    Raises
    ------
    TypeError
        If the `plugin_instance` is not a `PlugInFeature` instance
    """
    if isinstance(plugin_instance, PlugInFeature):
        # This check is necessary for situations where the PlugInFeature fails
        # between updating the `dclab.dfn` file and initialising the
        # AncillaryFeature
        if plugin_instance.feature_name in PlugInFeature.feature_names:
            PlugInFeature.feature_names.remove(plugin_instance.feature_name)
            feat_logic.feature_deregister(plugin_instance.feature_name)
        PlugInFeature.features.remove(plugin_instance)
    else:
        raise TypeError(f"Type {type(plugin_instance)} should be an instance "
                        f"of PlugInFeature. '{plugin_instance}' was given.")



import io
from os.path import join
import tempfile

import pytest

import dclab

from helper_methods import example_data_dict


def test_fcs_export():
    pytest.importorskip("fcswrite")
    keys = ["area_um", "deform", "time", "frame", "fl3_width"]
    ddict = example_data_dict(size=222, keys=keys)
    ds = dclab.new_dataset(ddict)

    edest = tempfile.mkdtemp()
    f1 = join(edest, "test.fcs")
    f2 = join(edest, "test_unicode.fcs")

    ds.export.fcs(f1, keys, override=True)
    ds.export.fcs(f2, [u"area_um", u"deform", u"time",
                       u"frame", u"fl3_width"], override=True)

    with io.open(f1, mode="rb") as fd:
        a1 = fd.read()

    with io.open(f2, mode="rb") as fd:
        a2 = fd.read()

    assert a1 == a2
    assert len(a1) != 0


def test_fcs_override():
    pytest.importorskip("fcswrite")
    keys = ["area_um", "deform", "time", "frame", "fl3_width"]
    ddict = example_data_dict(size=212, keys=keys)
    ds = dclab.new_dataset(ddict)

    edest = tempfile.mkdtemp()
    f1 = join(edest, "test.fcs")
    ds.export.fcs(f1, keys, override=True)
    try:
        ds.export.fcs(f1[:-4], keys, override=False)
    except OSError:
        pass
    else:
        raise ValueError("Should append .fcs and not override!")


def test_fcs_not_filtered():
    pytest.importorskip("fcswrite")
    keys = ["area_um", "deform", "time", "frame", "fl3_width"]
    ddict = example_data_dict(size=127, keys=keys)
    ds = dclab.new_dataset(ddict)

    edest = tempfile.mkdtemp()
    f1 = join(edest, "test.tsv")
    ds.export.fcs(f1, keys, filtered=False)



"""Statistics computation for RT-DC dataset instances"""

import numpy as np
import traceback as tb
import warnings

from . import definitions as dfn


class BadMethodWarning(UserWarning):
    pass


class Statistics(object):
    available_methods = {}

    def __init__(self, name, method, req_feature=False):
        """A helper class for computing statistics

        All statistical methods are registered in the dictionary
        `Statistics.available_methods`.
        """
        self.method = method
        self.name = name
        self.req_feature = req_feature
        Statistics.available_methods[name] = self

    def __call__(self, **kwargs):
        data = self._get_data(kwargs)
        if len(data) == 0:
            result = np.nan
        else:
            try:
                result = self.method(data)
            except BaseException:
                exc = tb.format_exc().replace("\n", "\n    | ")
                warnings.warn("Failed to compute {} for {}: {}".format(
                              self.name, kwargs["ds"].title, exc),
                              BadMethodWarning)
                result = np.nan
        return result

    def _get_data(self, kwargs):
        """Convenience wrapper to get statistics data"""
        if "ds" not in kwargs:
            raise ValueError("Keyword argument 'ds' missing.")

        ds = kwargs["ds"]

        if self.req_feature:
            if "feature" not in kwargs:
                raise ValueError("Keyword argument 'feature' missing.")
            return self.get_feature(ds, kwargs["feature"])
        else:
            return ds

    def get_feature(self, ds, feat):
        """Return filtered feature data

        The features are filtered according to the user-defined filters,
        using the information in `ds.filter.all`. In addition, all
        `nan` and `inf` values are purged.

        Parameters
        ----------
        ds: dclab.rtdc_dataset.RTDCBase
            The dataset containing the feature
        feat: str
            The name of the feature; must be a scalar feature
        """
        if ds.config["filtering"]["enable filters"]:
            x = ds[feat][ds.filter.all]
        else:
            x = ds[feat]
        bad = np.isnan(x) | np.isinf(x)
        xout = x[~bad]
        return xout


def flow_rate(ds):
    """Return the flow rate of an RT-DC dataset"""
    conf = ds.config["setup"]
    if "flow rate" in conf:
        return conf["flow rate"]
    else:
        return np.nan


def get_statistics(ds, methods=None, features=None):
    """Compute statistics for an RT-DC dataset

    Parameters
    ----------
    ds: dclab.rtdc_dataset.RTDCBase
        The dataset for which to compute the statistics.
    methods: list of str or None
        The methods wih which to compute the statistics.
        The list of available methods is given with
        `dclab.statistics.Statistics.available_methods.keys()`
        If set to `None`, statistics for all methods are computed.
    features: list of str
        Feature name identifiers are defined by
        `dclab.definitions.feature_exists`.
        If set to `None`, statistics for all scalar features
        available are computed.

    Returns
    -------
    header: list of str
        The header (feature + method names) of the computed statistics.
    values: list of float
        The computed statistics.
    """
    if methods is None:
        cls = list(Statistics.available_methods.keys())
        # sort the features in a usable way
        avm = Statistics.available_methods
        me1 = [m for m in cls if not avm[m].req_feature]
        me2 = [m for m in cls if avm[m].req_feature]
        methods = me1 + me2

    if features is None:
        features = ds.features_scalar
    else:
        features = [a.lower() for a in features]

    header = []
    values = []

    # First loop over all methods that do not require a feature
    for mt in methods:
        meth = Statistics.available_methods[mt]
        if not meth.req_feature:
            values.append(meth(ds=ds))
            header.append(mt)

    # To make sure that all methods are computed for each feature in a block,
    # we loop over all features. It would be easier to loop over the methods,
    # but the ordering of the resulting statistics would not be human-friendly.
    for ft in features:
        for mt in methods:
            meth = Statistics.available_methods[mt]
            if meth.req_feature:
                if ft in ds:
                    values.append(meth(ds=ds, feature=ft))
                else:
                    values.append(np.nan)
                label = dfn.get_feature_label(ft, rtdc_ds=ds)
                header.append(" ".join([mt, label]))

    return header, values


def mode(data):
    """Compute an intelligent value for the mode

    The most common value in experimental is not very useful if there
    are a lot of digits after the comma. This method approaches this
    issue by rounding to bin size that is determined by the
    Freedman–Diaconis rule.

    Parameters
    ----------
    data: 1d ndarray
        The data for which the mode should be computed.

    Returns
    -------
    mode: float
        The mode computed with the Freedman-Diaconis rule.
    """
    # size
    n = data.shape[0]
    # interquartile range
    iqr = np.percentile(data, 75)-np.percentile(data, 25)
    # Freedman–Diaconis
    bin_size = 2 * iqr / n**(1/3)

    if bin_size == 0:
        return np.nan

    # Add bin_size/2, because we want the center of the bin and
    # not the left corner of the bin.
    databin = np.round(data/bin_size)*bin_size + bin_size/2
    u, indices = np.unique(databin, return_inverse=True)
    mode = u[np.argmax(np.bincount(indices))]

    return mode


# Register all the methods
# Methods that require an axis
Statistics(name="Mean",   req_feature=True, method=np.average)
Statistics(name="Median", req_feature=True, method=np.median)
Statistics(name="Mode",   req_feature=True, method=mode)
Statistics(name="SD",     req_feature=True, method=np.std)
# Methods that work on RTDCBase
Statistics(name="Events",
           method=lambda mm: np.sum(mm.filter.all))
Statistics(name="%-gated",
           method=lambda mm: np.average(mm.filter.all)*100)
Statistics(name="Flow rate",
           method=lambda mm: flow_rate(mm))



"""Hooks for the LE-2D-FEM-19 dataset

https://doi.org/10.6084/m9.figshare.12155064.v4

See the individual functions for details.
"""
import pathlib

import matplotlib.pylab as plt
import numpy as np
from scipy import ndimage
from scipy.spatial import ConvexHull
from skimage import morphology

from dclab.features import emodulus

from .common import find_upper_intersection


def isoelastics_postprocess(lup, isoel):
    """Process extracted isoelastics"""
    new_isoel = []
    for ie in isoel:
        # remove noise at beginning
        if lup.featx == "area_um":
            xmin = 34.93
        elif lup.featx == "volume":
            xmin = 160
        else:
            raise NotImplementedError(f"Unexpected axis {lup.featx}!")
        ie = ie[ie[:, 0] > xmin]
        ie = ie[ie[:, 1] < .187]
        new_isoel.append(ie)
    return new_isoel


def get_analytical_volume_lut_2daxis():
    """Compute the volume-deformation analytical part of the LUT

    The data stored in LUT_analytical_linear-elastic_2Daxis.txt
    do not include the volume information (only the cross-sectional
    area of the deformed sphere). Since the linear elastic model
    means that volume is conserved, we can just compute the volume
    data by redoing the computations in the original Matlab script
    (CreateAvsChist_Loop_CH4.m). In addition, we have to crop the
    first 15 and the last data points which were manually removed
    to better fit in with the numerical values.
    """
    # analytical area_um-deform LUT
    ap = "LUT_analytical_linear-elastic_2Daxis.txt"
    lut_area = np.loadtxt(pathlib.Path(__file__).parent / ap)
    lut_volume = np.zeros_like(lut_area)
    lut_volume[:, 1] = lut_area[:, 1]
    lut_volume[:, 2] = lut_area[:, 2]

    # FEM data
    here = pathlib.Path(__file__).parent
    anap = here / "LUT_analytical_linear-elastic_2Daxis.txt"
    _, meta = emodulus.load_mtext(anap)
    assert meta["channel_width"] == 20
    assert meta["method"] == "analytical"
    assert meta["dimensionality"] == "2Daxis"

    # BEGIN MATLAB TRANSLATIONS
    # emodulus
    data1 = np.linspace(7.97**(-1/2), 28.43**(-1/2), 23)**(-2)
    nr_p = 100
    d = 20  # um
    # linear spaced area if assumed a sphere (spaced with square root)
    lambda_rand = np.linspace(0.01, 0.534, nr_p, endpoint=True)**(1/2)
    lambd = lambda_rand[np.abs(lambda_rand-0.5) < 0.5]  # 0<lambd<1
    # END MATLAB TRANSLATIONS
    # In the Matlab script, area in um is computed like this:
    # Area_unitless*1.094^2*d^2*lambda(i)^2/4;
    # (where Area_unitless is from the modeling computations).
    # - The unitless length is the radius of the cylindrical channel
    #   (lambd==1).
    # - The channel width d is always multiplied by the factor 1.094.
    radius = lambd * d/2 * 1.094
    volume = 4/3*np.pi * radius**3

    # The data stored in LUT_analytical_linear-elastic_2Daxis.txt does not
    # contain the full nr_p=100 points, but it was cropped manually *sigh*.
    # By manual inspection of of the highest emodulus isoelasticity line
    # and comparison with area=np.pi*radius**2, I am quite certain that the
    # first 15 data points and the last datapoint were cropped.
    volume = volume[15:-1]

    for emod in data1:
        eloc = np.abs(emod - lut_volume[:, 2]) < .01
        assert np.sum(eloc), "failed to find emodulus {}".format(emod)
        assert np.sum(eloc) == volume.size, "bad size emodulus {}".format(emod)
        lut_volume[eloc, 0] = volume

    return lut_volume


def lut_preprocess(lup):
    """Complement FEM LUT with analytical values, extrapolate, and crop it

    The LUT is complemented with analytical values from
    "LUT_analytical_linear-elastic_2Daxis.txt" for small deformation
    and area below 200um. The original FEM simulations did not cover that
    area, because of discretization problems (deformations smaller than
    0.005 could not be resolved).

    A sparse region for larger deformation is complemented with
    extrapolated isoelasticity lines (3rd order polynomial fit).
    See https://github.com/ZELLMECHANIK-DRESDEN/dclab/issues/191.

    The LUT is cropped at a maximum area of 290um^2. The reason is that
    the axis-symmetric model becomes inaccurate when the object boundary
    comes close to the channel walls (the actual flow profile in a
    rectangular cross-section channel is not anymore rotationally symmetric
    around the object). In addition, there have been numerical errors due
    to meshing if the area is above 290um^2.
    """
    if lup.featx == "area_um" and lup.featy == "deform":
        # Add additional values from analytical simulations and crop.
        lup.lut_raw = lut_preprocess_area_um_deform(lup)
        xcrop = 170
        xmax = 300
        num_isoel = 30
    elif lup.featx == "volume" and lup.featy == "deform":
        lup.lut_raw = lut_preprocess_volume_deform(lup)
        xcrop = 2300
        xmax = 3450
        num_isoel = 35
    else:
        raise NotImplementedError("Cannot process")

    # Perform linear interpolation until convex hull.
    # don't use small deformation values for isoelastics computation
    lut_iso = lup.lut_raw[
        np.logical_or(lup.lut_raw[:, 1] > 0.05,
                      lup.lut_raw[:, 0] > xcrop)
    ]

    emod, mask_sim = lup.map_lut_to_grid(lut_iso)
    # Find points that should not be in that 2D `emod` array.
    # (bad interpolation from convex vs raw enclosing polygon)
    # Apply a closing disk filter
    # Zero-pad the mask beforehand (otherwise the disk filter is not
    # properly applied to pixels at the boundary of the image)
    ds = 20  # disk closing size
    mask_padded = np.pad(mask_sim, ((ds, ds), (ds, ds)))
    mask_padded_disk = morphology.binary_closing(
        mask_padded, footprint=morphology.disk(ds))
    # Fill any holes (in case of sparse simulations)
    ndimage.binary_fill_holes(mask_padded_disk, output=mask_padded_disk)
    mask_disk = mask_padded_disk[ds:-ds, ds:-ds]

    emod[~mask_disk] = np.nan
    levels, contours, contours_px = lup.extract_isoelastics_from_grid(
        emod, lut=lut_iso, num=num_isoel)

    # Extend the LUT via extrapolation
    lut_new = lup.lut_raw
    ncs = []
    if lup.verbose:
        ax1 = plt.subplot(121, title="extrapolation region")
        ax1.imshow(1.*mask_disk + mask_sim)
        ax2 = plt.subplot(122, title="LUT extrapolation")
        plt.plot(lup.lut_raw[:, 0], lup.lut_raw[:, 1], ".")
    for lev, cc in zip(levels, contours):
        if np.isnan(lev) or len(cc) < 10:
            continue
        ip = np.polynomial.Polynomial.fit(cc[:, 0], cc[:, 1], 3)
        new_cc = np.zeros((20, 3))
        xrange = (cc[-1, 0], 1.2*cc[-1, 0])
        new_cc[:, 0] = np.linspace(xrange[0], xrange[1], 20)
        new_cc[:, 1] = ip(new_cc[:, 0])
        new_cc[:, 2] = lev
        inside = lup.points_in_convex_hull(new_cc[:, :2])
        # Add a few points outside the convex hull for isoelastics computation
        for ii in range(len(inside)):
            if not inside[ii]:
                inside[ii:ii+3] = True
                break
        parts = new_cc[inside]
        lut_new = np.concatenate((lut_new, parts))
        # Append boundary point
        try:
            sx, sy = find_upper_intersection(
                ip, lup.convex_hull, xrange=xrange)
        except BaseException:
            continue
        else:
            new_cc = np.concatenate((new_cc, [[sx, sy, lev]]))
            ncs.append(new_cc)

        if lup.verbose:
            xtest = np.linspace(cc[0, 0], 1.5*cc[-1, 0], 20, endpoint=True)
            ax2.plot(cc[:, 0], cc[:, 1], "o")
            ax2.plot(xtest, ip(xtest))
            ax2.plot(new_cc[:, 0], new_cc[:, 1], "x")

    if lup.verbose:
        ax2.plot(lup.convex_hull[:, 0], lup.convex_hull[:, 1])
        ax2.set_ylim(0, 0.2)
        ax2.set_xlim(0, xmax)
        plt.show()
    lup.lut_raw = lut_new


def lut_preprocess_area_um_deform(lup):
    """Complement FEM LUT with analytical values and crop it

    The LUT is complemented with analytical values from
    "LUT_analytical_linear-elastic_2Daxis.txt" for small deformation
    and area below 200um. The original FEM simulations did not cover that
    area, because of discretization problems (deformations smaller than
    0.005 could not be resolved).

    The LUT is cropped at a maximum area of 290um^2. The reason is that
    the axis-symmetric model becomes inaccurate when the object boundary
    comes close to the channel walls (the actual flow profile in a
    rectangular cross-section channel is not anymore rotationally symmetric
    around the object). In addition, there have been numerical errors due
    to meshing if the area is above 290um^2.
    """
    lut = lup.lut_raw
    ap = "LUT_analytical_linear-elastic_2Daxis.txt"
    print(f"...Post-Processing: Adding analytical LUT from {ap}.")
    # load analytical data
    lut_ana = np.loadtxt(pathlib.Path(__file__).parent / ap)
    lut = np.concatenate((lut, lut_ana))

    print("...Post-Processing: Cropping LUT at 290um^2.")
    lut = lut[lut[:, 0] < 290]
    lup.lut_raw = lut

    # Compute new convex hull of LUT
    hull = ConvexHull(lut[:, :2])
    lup.convex_hull = hull.points[hull.vertices, :]

    return lut


def lut_preprocess_volume_deform(lup):
    """Complement FEM LUT with analytical values and crop it

    The LUT is complemented with analytical values from
    "LUT_analytical_linear-elastic_2Daxis.txt" (which are extended
    to volume using the volume data used in the original Matlab script
    (see `get_analytical_part_2daxis`)) for small deformation and area
    below 200um. The original FEM simulations did not cover that
    area, because of discretization problems (deformations smaller than
    0.005 could not be resolved).

    The LUT is cropped at a maximum volume of 3200um^3. The reason is
    that the axis-symmetric model becomes inaccurate when the object
    boundary comes close to the channel walls (the actual flow profile
    in a rectangular cross-section channel is not anymore rotationally
    symmetric around the object). In addition, there have been numerical
    errors due to meshing if the area is above 290um^2.
    """
    print("...Post-Processing: Complementing analytical volume data.")
    # load analytical data
    lut_ana = get_analytical_volume_lut_2daxis()
    lut = np.concatenate((lup.lut_raw, lut_ana))

    print("...Post-Processing: Cropping LUT at 3200um^3.")
    # the analytical part (below) is anyhow below 200um^2
    # We cannot crop at 290um^2, because this will result in
    # a convex lut with interpolation taking place within it.
    # Converting the 290 to an equivalent sphere volume results
    # in a value outside the lut (3700 something). So we just
    # guess a value here:
    lut = lut[lut[:, 0] < 3200, :]
    lup.lut_raw = lut

    hull = ConvexHull(lut[:, :2])
    lup.convex_hull = hull.points[hull.vertices, :]

    return lut



"""Pixelation correction definitions"""

import numpy as np


def corr_deform_with_area_um(area_um, px_um=0.34):
    """Deformation correction for area_um-deform data

    The contour in RT-DC measurements is computed on a
    pixelated grid. Due to sampling problems, the measured
    deformation is overestimated and must be corrected.

    The correction formula is described in :cite:`Herold2017`.

    Parameters
    ----------
    area_um: float or ndarray
        Apparent (2D image) area in µm² of the event(s)
    px_um: float
        The detector pixel size in µm.

    Returns
    -------
    deform_delta: float or ndarray
        Error of the deformation of the event(s) that must be
        subtracted from `deform`.
        deform_corr = deform -  deform_delta
    """
    # A triple-exponential decay can be used to correct for pixelation
    # for apparent cell areas between 10 and 1250µm².
    # For 99 different radii between 0.4 μm and 20 μm circular objects were
    # simulated on a pixel grid with the pixel resolution of 340 nm/pix. At
    # each radius 1000 random starting points were created and the
    # obtained contours were analyzed in the same fashion as RT-DC data.
    # A convex hull on the contour was used to calculate the size (as area)
    # and the deformation.
    # The pixel size correction `pxscale` takes into account the pixel size
    # in the pixelation correction formula.
    pxscale = (.34 / px_um)**2
    offs = 0.0012
    exp1 = 0.020 * np.exp(-area_um * pxscale / 7.1)
    exp2 = 0.010 * np.exp(-area_um * pxscale / 38.6)
    exp3 = 0.005 * np.exp(-area_um * pxscale / 296)
    delta = offs + exp1 + exp2 + exp3

    return delta


def corr_deform_with_volume(volume, px_um=0.34):
    """Deformation correction for volume-deform data

    The contour in RT-DC measurements is computed on a
    pixelated grid. Due to sampling problems, the measured
    deformation is overestimated and must be corrected.

    The correction is derived in scripts/pixelation_correction.py.

    Parameters
    ----------
    volume: float or ndarray
        The "volume" feature (rotation of raw contour) [µm³]
    px_um: float
        The detector pixel size in µm.

    Returns
    -------
    deform_delta: float or ndarray
        Error of the deformation of the event(s) that must be
        subtracted from `deform`.
        deform_corr = deform -  deform_delta
    """
    pxscalev = (.34 / px_um)**3
    offs = 0.0013
    exp1 = 0.0172 * np.exp(-volume * pxscalev / 40)
    exp2 = 0.0070 * np.exp(-volume * pxscalev / 450)
    exp3 = 0.0032 * np.exp(-volume * pxscalev / 6040)
    delta = offs + exp1 + exp2 + exp3
    return delta


def get_pixelation_delta_pair(feat1, feat2, data1, data2, px_um=0.34):
    """Convenience function that returns pixelation correction pair"""
    # determine feature that defines abscissa
    feat_absc = feat1 if feat1 in ["area_um", "volume"] else feat2
    data_absc = data1 if feat_absc == feat1 else data2
    # first compute all the possible pixelation offsets
    delt1 = get_pixelation_delta(
        feat_corr=feat1,
        feat_absc=feat_absc,
        data_absc=data_absc,
        px_um=px_um)
    delt2 = get_pixelation_delta(
        feat_corr=feat2,
        feat_absc=feat_absc,
        data_absc=data_absc,
        px_um=px_um)
    return delt1, delt2


def get_pixelation_delta(feat_corr, feat_absc, data_absc, px_um=0.34):
    """Convenience function for obtaining pixelation correction

    Parameters
    ----------
    feat_corr: str
        Feature for which to compute the pixelation correction
        (e.g. "deform")
    feat_absc: str
        Feature with which to compute the correction (e.g. "area_um");
    data_absc: ndarray or float
        Corresponding data for `feat_absc`
    px_um: float
        Detector pixel size [µm]
    """
    if feat_corr == "deform" and feat_absc == "area_um":
        delt = corr_deform_with_area_um(data_absc, px_um=px_um)
    elif feat_corr == "circ" and feat_absc == "area_um":
        delt = -corr_deform_with_area_um(data_absc, px_um=px_um)
    elif feat_corr == "deform" and feat_absc == "volume":
        delt = corr_deform_with_volume(data_absc, px_um=px_um)
    elif feat_corr == "circ" and feat_absc == "volume":
        delt = -corr_deform_with_volume(data_absc, px_um=px_um)
    elif feat_corr == "area_um":
        # no correction for area
        delt = np.zeros_like(data_absc, dtype=float)
    elif feat_corr == "volume":
        # no correction for volume
        delt = np.zeros_like(data_absc, dtype=float)
    elif feat_corr == feat_absc:
        raise ValueError("Input feature names are identical!")
    else:
        raise KeyError(
            "No rule for feature '{}' with abscissa ".format(feat_corr)
            + "'{}'!".format(feat_absc))
    return delt



"""Young's modulus computation from data on DCOR

This example reproduces the lower right subplot of figure 10
in :cite:`Herold2017`. It illustrates how the Young's modulus
of elastic beads can be retrieved correctly (independent of
the flow rate, with correction for pixelation and shear-thinning)
using the area-deformation look-up table implemented in dclab
(right plot). For comparison, the flow-rate-dependent deformation
is also shown (left plot).

Note that this example uses the 'buyukurganci-2022' model for
computing the viscosity, which was introduced in dclab 0.48.0.

`The dataset <https://dcor.mpl.mpg.de/dataset/figshare-12721436-v1>`_
is loaded directly from `DCOR <https://dcor.mpl.mpg.de>`_ and thus
an active internet connection is required for this example.
"""
import dclab
import matplotlib.pylab as plt

# The dataset is also available on figshare
# (https://doi.org/10.6084/m9.figshare.12721436.v1), but we
# are accessing it through the DCOR API, because we do not
# have the time to download the entire dataset. The dataset
# name is figshare-12721436-v1. These are the resource IDs:
ds_loc = ["e4d59480-fa5b-c34e-0001-46a944afc8ea",
          "2cea205f-2d9d-26d0-b44c-0a11d5379152",
          "2cd67437-a145-82b3-d420-45390f977a90",
          ]
ds_list = []  # list of opened datasets
labels = []  # list of flow rate labels

# load the data
for loc in ds_loc:
    ds = dclab.new_dataset(loc)
    labels.append("{:.2f}".format(ds.config["setup"]["flow rate"]))
    # emodulus computation
    ds.config["calculation"]["emodulus lut"] = "LE-2D-FEM-19"
    ds.config["calculation"]["emodulus medium"] = ds.config["setup"]["medium"]
    ds.config["calculation"]["emodulus temperature"] = \
        ds.config["setup"]["temperature"]
    ds.config["calculation"]["emodulus viscosity model"] = 'buyukurganci-2022'
    # filtering
    ds.config["filtering"]["area_ratio min"] = 1.0
    ds.config["filtering"]["area_ratio max"] = 1.1
    ds.config["filtering"]["deform min"] = 0
    ds.config["filtering"]["deform max"] = 0.035
    # This option will remove "nan" events that appear in the "emodulus"
    # feature. If you are not working with DCOR, this might lead to a
    # longer computation time, because all available features are
    # computed locally. For data on DCOR, this computation already has
    # been done.
    ds.config["filtering"]["remove invalid events"] = True
    ds.apply_filter()
    # Create a hierarchy child for convenience reasons
    # (Otherwise we would have to do e.g. ds["deform"][ds.filter.all]
    # everytime we need to access a feature)
    ds_list.append(dclab.new_dataset(ds))

# plot
fig = plt.figure(figsize=(8, 4))

# box plot for deformation
ax1 = plt.subplot(121)
ax1.set_ylabel(dclab.dfn.get_feature_label("deform"))
data_deform = [di["deform"] for di in ds_list]
# Uncomment this line if you are not filtering invalid events (above)
# data_deform = [d[~np.isnan(d)] for d in data_deform]
bplot1 = ax1.boxplot(data_deform,
                     vert=True,
                     patch_artist=True,
                     labels=labels,
                     )

# box plot for Young's modulus
ax2 = plt.subplot(122)
ax2.set_ylabel(dclab.dfn.get_feature_label("emodulus"))
data_emodulus = [di["emodulus"] for di in ds_list]
# Uncomment this line if you are not filtering invalid events (above)
# data_emodulus = [d[~np.isnan(d)] for d in data_emodulus]
bplot2 = ax2.boxplot(data_emodulus,
                     vert=True,
                     patch_artist=True,
                     labels=labels,
                     )

# colors
colors = ["#0008A5", "#A5008D", "#A50100"]
for bplot in (bplot1, bplot2):
    for patch, color in zip(bplot['boxes'], colors):
        patch.set_facecolor(color)

# axes
for ax in [ax1, ax2]:
    ax.grid()
    ax.set_xlabel("flow rate [µL/s]")

plt.tight_layout()
plt.show()



import numpy as np

from ..meta_table import MetaTable


class DCORTables:
    def __init__(self, api):
        self.api = api
        self._tables_cache = None

    def __contains__(self, key):
        return key in self.keys()

    def __getitem__(self, key):
        return self._tables[key]

    def __iter__(self):
        for key in self.keys():
            yield key

    def __len__(self):
        return len(self._tables)

    def keys(self):
        return self._tables.keys()

    @property
    def _tables(self):
        if self._tables_cache is None:
            table_data = self.api.get(query="tables")
            # assemble the tables
            tables = {}
            for key in table_data:
                tables[key] = DCORTable(table_data[key])

            self._tables_cache = tables
        return self._tables_cache


class DCORTable(MetaTable):
    def __init__(self, table_content):
        self._columns, data = table_content
        self._tab_data = np.asarray(data)
        if self._columns is not None:
            # We have a rec-array (named columns)

            ds_dt = np.dtype({'names': self._columns,
                              'formats': [np.float64] * len(self._columns)})
            self._tab_data = np.rec.array(self._tab_data, dtype=ds_dt)

    def __array__(self, *args, **kwargs):
        return self._tab_data.__array__(*args, **kwargs)

    @property
    def meta(self):
        # TODO: Implement metadata sending from DCOR.
        return {}

    def has_graphs(self):
        return self._columns is not None

    def keys(self):
        return self._columns

    def __getitem__(self, key):
        return self._tab_data[key]



import dclab

from helper_methods import example_data_dict, retrieve_data

import pytest


def test_tsv_export(tmp_path):
    keys = ["area_um", "deform", "time", "frame", "fl3_width"]
    ddict = example_data_dict(size=222, keys=keys)
    ds = dclab.new_dataset(ddict)

    f1 = tmp_path / "test.tsv"
    f2 = tmp_path / "test_unicode.tsv"

    ds.export.tsv(f1, keys, override=True)
    ds.export.tsv(f2, [u"area_um", u"deform", u"time",
                       u"frame", u"fl3_width"], override=True)

    with f1.open("r", encoding="utf-8") as fd:
        a1 = fd.read()

    with f2.open("r", encoding="utf-8") as fd:
        a2 = fd.read()

    assert a1 == a2
    assert len(a1) != 0


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_tsv_export_ds_metadata(tmp_path):
    path = retrieve_data("fmt-hdf5_image-bg_2020.zip")
    ds = dclab.new_dataset(path)

    f1 = tmp_path / "test.tsv"

    ds.export.tsv(f1, ["area_um", "bright_avg"], override=True)

    lines = f1.read_text(encoding="utf-8").split("\n")

    for ll in [
        "# dc:experiment:date = 2020-10-23",
        "# dc:experiment:event count = 5",
        "# dc:experiment:run index = 1",
        "# dc:experiment:sample = background image example",
        "# dc:experiment:time = 10:37:15",
        "# dc:filtering:enable filters = True",
        "# dc:filtering:hierarchy parent = none",
        "# dc:filtering:limit events = 0",
        "# dc:filtering:polygon filters = []",
        "# dc:filtering:remove invalid events = False",
        "# dc:imaging:flash device = LED (ZMD L1)",
        "# dc:imaging:flash duration = 2.0",
        "# dc:imaging:frame rate = 2000.0",
        "# dc:imaging:pixel size = 0.34",
        "# dc:imaging:roi position x = 720",
        "# dc:imaging:roi position y = 512",
        "# dc:imaging:roi size x = 250",
        "# dc:imaging:roi size y = 80",
        "# dc:online_contour:bin area min = 10",
        "# dc:online_contour:bin kernel = 5",
        "# dc:online_contour:bin threshold = -6",
        "# dc:online_contour:image blur = 0",
        "# dc:online_contour:no absdiff = True",
        "# dc:online_filter:target event count = 5",
        "# dc:setup:channel width = 20.0",
        "# dc:setup:chip region = channel",
        "# dc:setup:flow rate = 0.16",
        "# dc:setup:flow rate sample = 0.04",
        "# dc:setup:flow rate sheath = 0.12",
        "# dc:setup:identifier = ZMDD-AcC-478cb7-dc924c",
    ]:
        assert ll in lines


def test_tsv_override(tmp_path):
    keys = ["area_um", "deform", "time", "frame", "fl3_width"]
    ddict = example_data_dict(size=212, keys=keys)
    ds = dclab.new_dataset(ddict)

    f1 = tmp_path / "test.tsv"
    ds.export.tsv(f1, keys, override=True)
    try:
        ds.export.tsv(f1.with_name(f1.stem), keys, override=False)
    except OSError:
        pass
    else:
        raise ValueError("Should append .tsv and not override!")


def test_tsv_not_filtered(tmp_path):
    keys = ["area_um", "deform", "time", "frame", "fl3_width"]
    ddict = example_data_dict(size=127, keys=keys)
    ds = dclab.new_dataset(ddict)

    f1 = tmp_path / "test.tsv"
    ds.export.tsv(f1, keys, filtered=False)



import abc


class MetaTable(abc.ABC):
    @abc.abstractmethod
    def __array__(self, *args, **kwargs):
        """Return array representation of the table"""

    @property
    @abc.abstractmethod
    def meta(self):
        """Return metadata of the table (e.g. graph colors)"""

    @abc.abstractmethod
    def has_graphs(self):
        """Return True when the table has key-based graphs"""

    @abc.abstractmethod
    def keys(self):
        """Return keys of the graphs, None if `not self.has_graphs()`"""

    @abc.abstractmethod
    def __getitem__(self, key):
        """Return a graph or otherwise part of the data array"""
        pass
